# TODO: use a more sophisticated system for this.
# When do quntities acquere units?
# - in config file?
# - when config is loaded?
# - when additional parameters are computed?

"""

The basic units are :
    kilogram                (kg)
    centimeter              (cm)
    microsecond             (us)
    positron charge         (e)
    Kelvin                  (kelvin)
    Mega electron-volt      (MeV)
    the amount of substance (mole)
    luminous intensity      (candela)
    radian                  (radian)
    steradian               (steradian)
"""

# mass
me = 9.109e-31
kg = 1.
g = 1.e-3*kg

# temperature
K = 1

# distance
km = 1.e5
m = 1.e2
cm = 1.
mm = 1.e-1
um = 1.e-4
nm = 1.e-7

# volume
mL = cm*cm*cm
L = 1.e3*mL

# time
s = 1.e6
ms = 1.e3
us = 1.
ns = 1.e-3

minute = 60*s
hour = 60*m

# charge
e = 1.
ke = 1.e3

C = 6.242e18*e

# potential
mV = 1.e-9
V = 1.e-6
kV = 1.e-3
MV = 1.

# energy
meV = 1.e-9
eV = 1.e-6
keV = 1.e-3
MeV = 1.
GeV = 1.e3
TeV = 1.e6

J = eV/e*C

# power
W = J/s

# angle
from math import pi
radians = 1.
degrees = 360./(2*pi)

def unit_parser(unit_string_expression):
    """
    unit_parser(unit_string_expression)

    Parser function for converting unit string expressions into a
    numerical representation based on the unit scheme defined in units.py
    
    Parameters
    ----------
    unit_string_expression : string
        String expression of units, in the form of "cm*cm/us/MeV".

    Returns
    -------
    unit_product : float
        Numerical representation of the unit string provided.
    
    """
    # units need to be in formats like:
    # cm*cm/s
    # no carrot notation! (right now)
    unit_words = []
    unit_power = [1]
    this_word = ''
    for char in unit_string_expression:
        if char == '*':
            unit_power.append(1)
            unit_words.append(this_word)
            this_word = ''
        elif char == '/':
            unit_power.append(-1)
            unit_words.append(this_word)
            this_word = ''
        elif char == ' ':
            continue
        else:
            this_word += char

    unit_words.append(this_word)

    unit_factor = []
    for this_word in unit_words:
        assert (this_word in globals()), this_word + " not found!"
        unit_factor.append(globals()[this_word])
        
    unit_product = 1
    for unit, power in zip(unit_factor, unit_power):
        unit_product *= pow(unit, power)

    return unit_product



import gampixpy
from gampixpy import units, mobility

import os
import yaml
import numpy as np
import torch

class Config (dict):
    """
    Config(config_filename)

    Initialize a new config dict from a yaml file.  This class serves
    as the parent class for specialized config classes for detector,
    physics, and readout parameter settings.  This class defines the
    methods for reading specifications from the input, resolving units
    to the internal unit scheme, and computing derived parameters,
    returning a dict-like object containing all parameters and their values.

    Parameters
    ----------
    config_filename : path_like
        A string or os.path-like object pointing to a yaml file containing
        definitions for parameters.

    Returns
    -------
    out : Config
        A dict-like object containing input and derived parameters.

    See Also
    --------
    DetectorConfig : Sub-class for parsing parameters for detector geometry
                     and steering.
    PhysicsConfig : Sub-class for parsing parameters for physics processes
                    (recombination, charge mobility, etc.)
    ReadoutConfig : Sub-class for parsing parameters for readout details.

    Examples
    --------
    >>> c = Config('path/to/config.yaml')

    """
    def __init__(self, config_filename):
        self.config_filename = config_filename

        self._parse_config()
        self._compute_derived_parameters()
        
    def _parse_config(self):
        # parse a config (yaml) file and store values
        # internally as a dict
        with open(self.config_filename) as config_file:
            self.update(yaml.load(config_file, Loader = yaml.FullLoader).items())

        # where quantities with units are specified,
        # resolve them to the internal unit system
        self.update(self._resolve_units(self))

    def _resolve_units(self, sub_dict):
        if 'value' in sub_dict and 'unit' in sub_dict:
            numerical_value = sub_dict['value']
            unit = units.unit_parser(sub_dict['unit'])
            resolved_dict = numerical_value*unit
        else:
            resolved_dict = {}
            for key, value in sub_dict.items():
                if type(value) == dict:
                    resolved_dict[key] = self._resolve_units(value)
                else:
                    resolved_dict[key] = value

        return resolved_dict

class DetectorConfig (Config):
    """
    DetectorConfig(config_filename)

    Initialize a new detector config dict from a yaml file.  This class
    reads specifications from the input, resolves units to the internal
    unit scheme, and computes derived parameters, returning a dict-like
    object containing all parameters and their values.

    Parameters
    ----------
    config_filename : path_like
        A string or os.path-like object pointing to a yaml file containing
        definitions for detector geometry parameters.

    Returns
    -------
    out : DetectorConfig
        A dict-like object containing input and derived parameters for
        detector geometry.

    See Also
    --------
    Config : Parent config class which does not computation of derived
             parameters.
    PhysicsConfig : Similar config class for parsing parameters for
                    physics processes (recombination, charge mobility,
                    etc.)
    ReadoutConfig : Similar config class for parsing parameters for
                    readout details.

    Examples
    --------
    >>> dc = DetectorConfig('path/to/config.yaml')

    """
    def _compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        # calculate the span of the anode in TPC coordinates
        # based on the provided center
        
        # need to handle the case where the pitch doesn't evenly divide the span of the anode
        for volume_name, volume_dict in self['drift_volumes'].items():
            anode_center = torch.tensor([volume_dict['anode_center']['x'],
                                         volume_dict['anode_center']['y'],
                                         volume_dict['anode_center']['z']]).float()

            vertical_axis = torch.tensor([volume_dict['anode_vertical']['x'],
                                          volume_dict['anode_vertical']['y'],
                                          volume_dict['anode_vertical']['z']]).float()
            vertical_axis = vertical_axis/torch.sqrt(torch.inner(vertical_axis, vertical_axis))

            drift_axis = torch.tensor([volume_dict['drift_direction']['x'],
                                       volume_dict['drift_direction']['y'],
                                       volume_dict['drift_direction']['z']]).float()
            drift_axis = drift_axis/torch.sqrt(torch.inner(drift_axis, drift_axis))
            # drift_axis = drift_axis/np.linalg.norm(drift_axis)

            message = "anode vertical axis is not perpendicular to given drift axis!"
            assert torch.inner(vertical_axis, drift_axis) == 0, message

            horizontal_axis = torch.linalg.cross(vertical_axis,
                                                 drift_axis)

            half_span_horizontal = horizontal_axis*volume_dict['anode_span']['width']/2
            half_span_vertical = vertical_axis*volume_dict['anode_span']['height']/2

            anode_corners = [anode_center - half_span_horizontal - half_span_vertical,
                             anode_center - half_span_horizontal + half_span_vertical,
                             anode_center + half_span_horizontal - half_span_vertical,
                             anode_center + half_span_horizontal + half_span_vertical,
                             ]

            depth_span = drift_axis*volume_dict['depth']
            cathode_corners = [anode_center - half_span_horizontal - half_span_vertical + depth_span,
                               anode_center - half_span_horizontal + half_span_vertical + depth_span,
                               anode_center + half_span_horizontal - half_span_vertical + depth_span,
                               anode_center + half_span_horizontal + half_span_vertical + depth_span,
                               ]

            anode_mask = [1, 1, 1, 1, 0, 0, 0, 0]
            # hard-coded adjacency based on the definition of the corners above
            connectivity = torch.tensor([[0, 1, 1, 0, 1, 0, 0, 0],
                                         [1, 0, 0, 1, 0, 1, 0, 0],
                                         [1, 0, 0, 1, 0, 0, 1, 0],
                                         [0, 1, 1, 0, 0, 0, 0, 1],
                                         [1, 0, 0, 0, 0, 1, 1, 0],
                                         [0, 1, 0, 0, 1, 0, 0, 1],
                                         [0, 0, 1, 0, 1, 0, 0, 1],
                                         [0, 0, 0, 1, 0, 1, 1, 0]], dtype = bool)

            volume_dict.update({'anode_center': anode_center,
                                'anode_vertical': vertical_axis,
                                'anode_horizontal': horizontal_axis,
                                'drift_axis': drift_axis,
                                'anode_corners': anode_corners,
                                'cathode_corners': cathode_corners,
                                'corners': torch.stack(anode_corners + cathode_corners),
                                'connectivity': connectivity,
                                })
        
        # print (self['drift_volumes'])
        # self['tile_volume_edges'] = (np.linspace(self['anode']['x_lower_bound'],
        #                                          self['anode']['x_upper_bound'],
        #                                          self['n_tiles_x']+1),
        #                              np.linspace(self['anode']['y_lower_bound'],
        #                                          self['anode']['y_upper_bound'],
        #                                          self['n_tiles_y']+1))
        return

class PhysicsConfig (Config):
    """
    PhysicsConfig(config_filename)

    Initialize a new pysics config dict from a yaml file.  This class
    reads specifications from the input, resolves units to the internal
    unit scheme, and computes derived parameters, returning a dict-like
    object containing all parameters and their values.

    Parameters
    ----------
    config_filename : path_like
        A string or os.path-like object pointing to a yaml file containing
        definitions for physics parameters.

    Returns
    -------
    out : PhysicsConfig
        A dict-like object containing input and derived parameters for
        physics processes.

    See Also
    --------
    Config : Parent config class which does not computation of derived
             parameters.
    DetectorConfig : Similar config class for parsing parameters for
                     detector geometry and steering.
    ReadoutConfig : Similar config class for parsing parameters for
                    readout details.

    Examples
    --------
    >>> pc = PhysicsConfig('path/to/config.yaml')

    """
    def _compute_derived_parameters(self):
        mobility_model = mobility.MobilityModel(self)
        self['charge_drift'].update(mobility_model.compute_parameters())

        return

class ReadoutConfig (Config):
    """
    ReadoutConfig(config_filename)

    Initialize a new readout config dict from a yaml file.  This class
    reads specifications from the input, resolves units to the internal
    unit scheme, and computes derived parameters, returning a dict-like
    object containing all parameters and their values.

    Parameters
    ----------
    config_filename : path_like
        A string or os.path-like object pointing to a yaml file containing
        definitions for readout parameters.

    Returns
    -------
    out : ReadoutConfig
        A dict-like object containing input and derived parameters for
        readout electronics simulation.

    See Also
    --------
    Config : Parent config class which does not computation of derived
             parameters.
    DetectorConfig : Similar config class for parsing parameters for
                     detector geometry and steering.
    PhysicsConfig : Similar config class for parsing parameters for
                    physics processes (recombination, charge mobility,
                    etc.)

    Examples
    --------
    >>> rc = ReadoutConfig('path/to/config.yaml')

    """
    def _compute_derived_parameters(self):
        # compute any required parameters from
        # those specified in the YAML

        # self['n_pixels_x'] = int((self['anode']['x_upper_bound'] - self['anode']['x_lower_bound'])/self['pixels']['pitch'])
        # self['n_pixels_y'] = int((self['anode']['y_upper_bound'] - self['anode']['y_lower_bound'])/self['pixels']['pitch'])

        # self['n_tiles_x'] = int((self['anode']['x_upper_bound'] - self['anode']['x_lower_bound'])/self['coarse_tiles']['pitch'])
        # self['n_tiles_y'] = int((self['anode']['y_upper_bound'] - self['anode']['y_lower_bound'])/self['coarse_tiles']['pitch'])

        return

default_detector_params = DetectorConfig(os.path.join(gampixpy.__path__[0],
                                                      'detector_config',
                                                      'default.yaml'))
default_physics_params = PhysicsConfig(os.path.join(gampixpy.__path__[0],
                                                    'physics_config',
                                                    'default.yaml'))
default_readout_params = ReadoutConfig(os.path.join(gampixpy.__path__[0],
                                                    'readout_config',
                                                    'default.yaml'))

# far_detector_params = DetectorConfig(os.path.join(gampixpy.__path__[0],
#                                                       'detector_config',
#                                                       'far_detector.yaml'))
# far_readout_params = ReadoutConfig(os.path.join(gampixpy.__path__[0],
#                                                     'readout_config',
#                                                     'far_detector.yaml'))



import h5py
import numpy as np

from gampixpy.readout_objects import coarse_tile_dtype, pixel_dtype
from gampixpy.input_parsing import meta_dtype

class OutputManager:
    """
    OutputManager(output_filename)

    Manager class for writing simulation output to a file on disk.  This
    class defines how to handle the HDF5 datasets, as well as how to add
    entries to a dataset on disk sequentially.

    Attributes
    ----------
    output_filename : string or os.path-like
        Path to output destination on disk.
    outfile : h5py.File
        Opened file handle for the output file.
    n_tracks : int
        Number of tracks written to the output file so far.
    
    """
    def __init__(self, output_filename):
        self.output_filename = output_filename

        self.outfile = h5py.File(output_filename, 'w')

        self.outfile.create_dataset('coarse_hits',
                                    shape = (0,),
                                    dtype = coarse_tile_dtype,
                                    maxshape = (None,))
        self.outfile.create_dataset('pixel_hits',
                                    shape = (0,),
                                    dtype = pixel_dtype,
                                    maxshape = (None,))
        self.outfile.create_dataset('meta',
                                    shape = (0,),
                                    dtype = meta_dtype,
                                    maxshape = (None,))
        self.n_tracks = 0 # track how many tracks have been written so far

    def add_entry(self, track, meta, event_id = None):
        """
        om.add_entry(track, meta, event_id = None)

        Add a new row to the output datasets

        Parameters
        ----------
        track : Track object
            A track which has been fully simulated and has coarse tile and
            pixel hits populated.
        meta : array-like[float]
            Metadata array for this entry.  Metadata dtype is described in
            input_parsing.meta_dtype.
        event_id : int
            Index to assign this image in the output dataset.  If not
            specified, use the current value of the internal counter.
        
        """
        self._add_track(track, event_id)
        self._add_meta(meta, event_id)

        self.n_tracks += 1
        
    def _add_track(self, track, event_id = None):
        coarse_tile_sample_array, pixel_sample_array = track.to_array()

        if event_id:
            coarse_tile_sample_array[:]['event id'] = event_id
            pixel_sample_array[:]['event id'] = event_id
        else:
            coarse_tile_sample_array[:]['event id'] = self.n_tracks
            pixel_sample_array[:]['event id'] = self.n_tracks
        
        n_coarse_hits = coarse_tile_sample_array.shape[0]
        n_coarse_hits_prev = self.outfile['coarse_hits'].shape[0]

        n_pixel_hits = pixel_sample_array.shape[0]
        n_pixel_hits_prev = self.outfile['pixel_hits'].shape[0]

        self.outfile['coarse_hits'].resize((n_coarse_hits+n_coarse_hits_prev,))
        self.outfile['coarse_hits'][n_coarse_hits_prev:] = coarse_tile_sample_array

        self.outfile['pixel_hits'].resize((n_pixel_hits+n_pixel_hits_prev,))
        self.outfile['pixel_hits'][n_pixel_hits_prev:] = pixel_sample_array

    def _add_meta(self, meta, event_id):
        if event_id:
            meta[:]['event id'] = event_id
        else:
            meta[:]['event id'] = self.n_tracks
        
        self.outfile['meta'].resize((self.n_tracks + 1,))
        self.outfile['meta'][self.n_tracks:] = meta



from gampixpy.config import default_detector_params, default_physics_params, default_readout_params
from gampixpy.readout_objects import PixelSample, CoarseGridSample
from gampixpy.coordinates import CoordinateManager

import numpy as np
import torch
import torchist

class ReadoutModel:
    """
    ReadoutModel(readout_config)

    Parent class for readout models.  This class defines the process of
    taking drifted point clouds and forming current series for each coarse
    tile and pixel in the volume (for some defined tile/pixel response model).

    Attributes
    ----------
    readout_config : ReadoutConfig object
        Config object containing specifications for tile and pixel size, gaps,
        threshold, noise, etc.
    clock_start_time : float
        timestamp of the earliest arriving charge bundle.  This serves as the
        first clock tick of the event readout.
    coarse_tile_hits : array-like[CoarseTileHit]
        Array of coarse tile hits populated by tile_hit_finding method
    fine_pixel_hits : array-like[CoarseTileHit]
        Array of pixel hits populated by pixel_hit_finding method

    Notes
    -----
    tile_hit_finding and pixel_hit_finding methods are implemented by sub-classes.

    See Also
    --------
    GAMPixModel : Sub-class implementing a GAMPix-like hit-finding scheme.
    LArPixModel : Sub-class implementing a LArPix-like hit-finding scheme.    

    """
    def __init__(self,
                 readout_config = default_readout_params,
                 physics_config = default_physics_params,
                 detector_config = default_detector_params):
        self.readout_config = readout_config
        self.physics_config = physics_config
        self.detector_config = detector_config

        self.coordinate_manager = CoordinateManager(detector_config)

        self.clock_start_time = 0

    def electronics_simulation(self, track, verbose = True, **kwargs):
        """
        readout.electronics_simulation(track,
                                       verbose = True,
                                       **kwargs)

        Steering function for the individual steps of electronics simulation.

        Parameters
        ----------
        track : Track object
            Input event representation containing a populated drifted_track.
        verbose : bool
            Print update messages as simulation progresses through each stage.
        
        """
        try:
            self.clock_start_time = torch.min(track.drifted_track['time'])
        except RuntimeError:
            print ("Encountered empty event!")
            message = """This is most often because the event lies totally outside of the specified detector volume"""
            print (message)
            print ("skipping...")

        if verbose:
            print ("simulating coarse grid...")
        # do coarse grid binning/time series formation
        coarse_tile_timeseries = self.tile_current_builder(track, **kwargs)
        if verbose:
            print ("coarse time series built")
            print ("coarse hit finding...")
        # find hits on coarse grid
        self.coarse_tile_hits = self.tile_hit_finding(track, coarse_tile_timeseries, **kwargs)

        if verbose:
            print ("simulating fine grid...")
        fine_pixel_timeseries = self.pixel_current_builder(track, self.coarse_tile_hits, **kwargs)
        if verbose:
            print ("pixel time series built")
            print ("pixel hit finding...")
        # find hits on fine pixels
        self.fine_pixel_hits = self.pixel_hit_finding(track, fine_pixel_timeseries, **kwargs)

        if verbose:
            print ("found", len(track.coarse_tiles_samples), "coarse tile hits")
            print ("found", len(track.pixel_samples), "pixel hits")
        
        return 

    def point_sample_tile_current(self, tile_coord, track, sample_mask):
        """
        readout.point_sample_tile_current(tile_coord,
                                          track,
                                          sample_mask)

        From the array of arrival times and charges, build a stacked array of
        charge response and time ticks.

        Parameters
        ----------
        tile_coord : tuple(int, float, float)
            The (i_tpc, x, y) (anode) coordinates of a given tile.
        track : Track object
            A track which has already had drift applied.
        sample_mask : array-like[bool]
            A boolean mask specifying which point sample charges will have
            a non-negligible induction on the tile of interest.

        Returns
        -------
        stacked_induced_charge : array-like
            A stacked array of timestamps and instantaneous charge.  Shape: (2, N_samples).

        """
        # for now, return the entire charge with the time of arrival
        # really, should return a sparse timeseries informed by FEM
        # probably loaded from a table and interpolated
        # shape will be (2, N_point_samples, N_time_samples)
        # the first axis is (timestamps, charge mass)
        # the timestamps represent the time (relative to event t0)
        # at the beginning of a small interval (ideally, much smaller than the clock)
        # and charge is the integrated current across that interval

        # sum(result[1,:,:], dim = -1) == track.drifted_track['charge']
        
        position = track.drifted_track['position'][sample_mask]
        tpc_index = track.tpc_track['TPC_index'][sample_mask]
        charge = track.drifted_track['charge'][sample_mask]
        time = track.drifted_track['time'][sample_mask]
        pitch = self.readout_config['coarse_tiles']['pitch']
        
        lands_on_tile_of_interest = position[:,0] - tile_coord[1] > -0.5*pitch
        lands_on_tile_of_interest *= position[:,0] - tile_coord[1] < 0.5*pitch
        lands_on_tile_of_interest *= position[:,1] - tile_coord[2] > -0.5*pitch
        lands_on_tile_of_interest *= position[:,1] - tile_coord[2] < 0.5*pitch

        induced_charge = torch.where(lands_on_tile_of_interest,
                                     charge,
                                     torch.zeros(1),
                                     )
        
        return torch.stack((time, induced_charge))[:,:,None] 
    
    def compose_tile_currents(self, sparse_current_series):
        """
        readout.compose_tile_currents(sparse_current_series)

        Combine multiple sparse timeseries into a single dense timeseries.

        Parameters
        ----------
        sparse_current_series : array-like
            Stacked array (shape: (2, N_samples)) containing sparse timeseries data.
            Expected array is output from readout.point_sample_tile_current.
        
        Returns
        -------
        clock_ticks : array-like
            Array of timestamps at the beginning of each clock cycle.
        induced_charge : array-like
            Array of integrated charge induced on pixel within each clock cycle.
        
        """
        try: # there is a bug where sparse_current_series is sometimes empty            
            last_charge_arrival_time = torch.max(sparse_current_series[0,:,:])
            # when there is only one charge sample in a coarse cell's field
            # and it is also the earliest charge sample, n_clock_ticks is 0
            # so, add an extra clock tick to be safe
            n_clock_ticks = torch.ceil((last_charge_arrival_time - self.clock_start_time)/self.readout_config['coarse_tiles']['clock_interval']).int() + 1 
            
            arrival_time_bin_edges = torch.linspace(self.clock_start_time,
                                                    self.clock_start_time + n_clock_ticks*self.readout_config['coarse_tiles']['clock_interval'],
                                                    n_clock_ticks + 1,
                                                    )

            # find the induced charge which falls into each clock bin
            induced_charge = torchist.histogram(sparse_current_series[0,:,:],
                                                weights = sparse_current_series[1,:,:],
                                                edges = arrival_time_bin_edges)
        except RuntimeError:
            arrival_time_bin_edges = torch.tensor([self.clock_start_time,
                                                   self.clock_start_time + self.readout_config['coarse_tiles']['clock_interval'],
                                                   ])
            induced_charge = torch.zeros_like(arrival_time_bin_edges)

        return arrival_time_bin_edges[:-1], induced_charge
        
    def tile_receptive_field(self, tile_coord, track, n_neighbor_tiles = 0, **kwargs):
        """
        readout.tile_receptive_field(tile_coord,
                                     track,
                                     n_neightbor_tiles = 0)
        
        For each tile, select the track samples within the tile's receptive field.

        Parameters
        ----------
        tile_coord : tuple(int, float, float)
            The (i_tpc, x, y) (anode) coordinates of a given tile.
        track : Track object
            A track which has already had drift applied.
        n_neighbor_tiles : int
            The number of adjacent tiles to include (1 means include the 8 adjacent
            and semi-adjacent tiles surrounding the provided tile coord).

        Returns
        -------
        sample_mask : array-like[bool]
            A boolean array with the same shape as the number of point samples
            contained in the `track` argument.
        
        """

        position = track.drifted_track['position']
        pitch = self.readout_config['coarse_tiles']['pitch']

        sample_mask = position[:,0] - tile_coord[1] > -(n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,0] - tile_coord[1] <= (n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,1] - tile_coord[2] > -(n_neighbor_tiles + 0.5)*pitch
        sample_mask *= position[:,1] - tile_coord[2] <= (n_neighbor_tiles + 0.5)*pitch

        return sample_mask 

    def tile_current_builder(self, track, **kwargs):
        """
        readout.tile_current_builder(track, **kwargs)

        Build the current timeseries for each tile.

        Parameters
        ----------
        track : Track obect
            Input event representation containing a populated drited_track.

        Returns
        -------
        coarse_grid_timeseries : dict{tuple: array-like}
            Dictionary of charge timeseries for tiles in detector which measure
            some charge.
        
        """
        coarse_grid_timeseries = {}
        pitch = self.readout_config['coarse_tiles']['pitch']
        for volume_name in self.detector_config['drift_volumes'].keys():
            i_tpc = self.coordinate_manager.volume_to_index[volume_name]
            tpc_mask = track.tpc_track['TPC_index'] == i_tpc
            
            min_tile = torch.tensor([-0.5*self.detector_config['drift_volumes'][volume_name]['anode_span']['width'],
                                     -0.5*self.detector_config['drift_volumes'][volume_name]['anode_span']['height'],
                                     ])
            tile_ind = torch.floor(torch.div(track.drifted_track['position'][tpc_mask,:2] - min_tile, pitch)).int()

            unique_tile_indices = torch.unique(tile_ind, dim = 0)

            for this_tile_ind in unique_tile_indices:
                tile_center = min_tile + (this_tile_ind + 0.5*torch.ones(2))*pitch
                tile_coord = (i_tpc,
                              round(float(tile_center[0]), 3),
                              round(float(tile_center[1]), 3))

                sample_mask = self.tile_receptive_field(tile_coord, track)
                tile_sample_current_series = self.point_sample_tile_current(tile_coord,
                                                                            track,
                                                                            sample_mask)
                tile_current_series = self.compose_tile_currents(tile_sample_current_series)
                coarse_grid_timeseries[tile_coord] = tile_current_series

        return coarse_grid_timeseries

    def point_sample_pixel_current(self, pixel_coord, track, sample_mask):
        """
        readout.point_sample_pixel_current(tile_coord,
                                           track,
                                           sample_mask)

        Build a sparse timeseries for each input position.  Positions are
        relative to the pixel center.

        Parameters
        ----------
        pixel_coord : tuple(float, float)
            The (x, y) (anode) coordinates of a given pixel.
        track : Track object
            A track which has already had drift applied.
        sample_mask : array-like[bool]
            A boolean mask specifying which point sample charges will have
            a non-negligible induction on the pixel of interest.

        Returns
        -------
        stacked_induced_charge : array-like
            A stacked array of timestamps and instantaneous charge.  Shape: (2, N_samples).

        """
        # for now, return the entire charge with the time of arrival
        # really, should return a sparse timeseries informed by FEM
        # probably loaded from a table and interpolated
        # shape will be (2, N_point_samples, N_time_samples)
        # the first axis is (timestamps, charge mass)
        # the timestamps represent the time (relative to event t0)
        # at the beginning of a small interval (ideally, much smaller than the clock)
        # and charge is the integrated current across that interval

        # sum(result[1,:,:], dim = -1) == track.drifted_track['charge']
        
        position = track.drifted_track['position'][sample_mask]
        charge = track.drifted_track['charge'][sample_mask]
        time = track.drifted_track['time'][sample_mask]
        pitch = self.readout_config['pixels']['pitch']

        lands_on_pixel_of_interest = position[:,0] - pixel_coord[0] > -0.5*pitch
        lands_on_pixel_of_interest *= position[:,0] - pixel_coord[0] < 0.5*pitch
        lands_on_pixel_of_interest *= position[:,1] - pixel_coord[1] > -0.5*pitch
        lands_on_pixel_of_interest *= position[:,1] - pixel_coord[1] < 0.5*pitch

        induced_charge = torch.where(lands_on_pixel_of_interest,
                                     charge,
                                     torch.zeros(1),
                                     )
        return torch.stack((time, induced_charge))[:,:,None] 

    def compose_pixel_currents(self, sparse_current_series, coarse_cell_hit):
        """
        readout.compose_pixel_currents(sparse_current_series, coarse_cell_hit)

        Combine multiple sparse timeseries into a single dense timeseries.

        Parameters
        ----------
        sparse_current_series : array-like
            Stacked array (shape: (2, N_samples)) containing sparse timeseries data.
            Expected array is output from readout.point_sample_tile_current.
        coarse_cell_hit : CoarseGridSample object
            The coarse cell hit inside which this pixel lies.
        
        Returns
        -------
        clock_ticks : array-like
            Array of timestamps at the beginning of each clock cycle.
        induced_charge : array-like
            Array of integrated charge induced on pixel within each clock cycle.
        
        """
        cell_clock_start = coarse_cell_hit.coarse_measurement_time
        cell_clock_end = coarse_cell_hit.coarse_measurement_time + self.readout_config['coarse_tiles']['clock_interval']*self.readout_config['coarse_tiles']['integration_length']
        
        n_clock_ticks = int((cell_clock_end - cell_clock_start)/self.readout_config['pixels']['clock_interval'])
        arrival_time_bin_edges = torch.linspace(cell_clock_start,
                                                cell_clock_end,
                                                n_clock_ticks + 1,
                                                )

        # find the induced charge which falls into each clock bin
        induced_charge = torchist.histogram(sparse_current_series[0,:,:],
                                            weights = sparse_current_series[1,:,:],
                                            edges = arrival_time_bin_edges)

        return arrival_time_bin_edges[:-1], induced_charge

    def pixel_receptive_field(self, pixel_coord, track, n_neighbor_pixels = 0, **kwargs):
        """
        readout.tile_receptive_field(pixel_coord,
                                     track,
                                     n_neightbor_pixels = 0)
        
        For each tile, select the track samples within the pixel's receptive field

        Parameters
        ----------
        pixel_coord : tuple(float, float)
            The (x, y) (anode) coordinates of a given pixel.
        track : Track object
            A track which has already had drift applied.
        n_neighbor_pixels : int
            The number of adjacent pixels to include (1 means include the 8 adjacent
            and semi-adjacent pixels surrounding the provided pixel coord).

        Returns
        -------
        sample_mask : array-like[bool]
            A boolean array with the same shape as the number of point samples
            contained in the `track` argument.
        
        """

        position = track.drifted_track['position']
        pitch = self.readout_config['pixels']['pitch']

        sample_mask = position[:,0] - pixel_coord[0] > -(n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,0] - pixel_coord[0] <= (n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,1] - pixel_coord[1] > -(n_neighbor_pixels + 0.5)*pitch
        sample_mask *= position[:,1] - pixel_coord[1] <= (n_neighbor_pixels + 0.5)*pitch

        return sample_mask 
    
    def pixel_current_builder(self, track, coarse_tile_hits, **kwargs):
        """
        readout.pixel_current_builder(track, coarse_tile_hits, **kwargs)

        Build the current timeseries for each pixel.

        Parameters
        ----------
        track : Track obect
            Input event representation containing a populated drited_track.
        coarse_tile_hits : list[CoarseGridSample]
            List of found coarse hits.

        Returns
        -------
        pixel_timeseries : dict{tuple: array-like}
            Dictionary of charge timeseries for pixels in detector which measure
            some charge.
        
        """
        pixel_timeseries = {}
        
        tile_pitch = self.readout_config['coarse_tiles']['pitch']
        pixel_pitch = self.readout_config['pixels']['pitch']
        
        for this_coarse_hit in coarse_tile_hits:

            cell_tpc = this_coarse_hit.coarse_cell_tpc # may change
            cell_center_xy = this_coarse_hit.coarse_cell_pos # may change
            cell_trigger_t = this_coarse_hit.coarse_measurement_time
            
            x_bounds = [cell_center_xy[0] - 0.5*tile_pitch,
                        cell_center_xy[0] + 0.5*tile_pitch]
            y_bounds = [cell_center_xy[1] - 0.5*tile_pitch,
                        cell_center_xy[1] + 0.5*tile_pitch]
            t_bounds = [cell_trigger_t,
                        cell_trigger_t + self.readout_config['coarse_tiles']['clock_interval']*self.readout_config['coarse_tiles']['integration_length']]

            in_cell_mask = track.drifted_track['position'][:,0] >= x_bounds[0]
            in_cell_mask *= track.drifted_track['position'][:,0] < x_bounds[1]
            in_cell_mask *= track.drifted_track['position'][:,1] >= y_bounds[0]
            in_cell_mask *= track.drifted_track['position'][:,1] < y_bounds[1]
            in_cell_mask *= track.drifted_track['time'] >= t_bounds[0]
            in_cell_mask *= track.drifted_track['time'] < t_bounds[1]
            in_cell_mask *= track.tpc_track['TPC_index'] == cell_tpc

            in_cell_positions = track.drifted_track['position'][in_cell_mask]
            in_cell_charges = track.drifted_track['charge'][in_cell_mask]
            
            min_pixel = torch.tensor([x_bounds[0],
                                      y_bounds[0],
                                      ])
            n_pixels_x = int((x_bounds[1] - x_bounds[0])/pixel_pitch)
            n_pixels_y = int((y_bounds[1] - y_bounds[0])/pixel_pitch)
            pixel_volume_edges = (torch.linspace(x_bounds[0], x_bounds[1], n_pixels_x+1),
                                  torch.linspace(y_bounds[0], y_bounds[1], n_pixels_y+1))
            
            z_series = in_cell_positions[:,2]
            t_series = track.drifted_track['time'][in_cell_mask]
            charge_series = in_cell_charges

            # generate a unique id for each pixel within this coarse hit
            # so that hits from 
            pixel_ind = torch.floor(torch.div(in_cell_positions[:,[0, 1]] - min_pixel, pixel_pitch)).int()
            unique_pixel_indices = torch.unique(pixel_ind, dim = 0)
            
            for this_pixel_ind in unique_pixel_indices:
                pixel_center = torch.tensor([pixel_volume_edges[i][this_pixel_ind[i]] + 0.5*pixel_pitch
                                             for i in range(2)])
                pixel_coord = (round(float(pixel_center[0]), 3),
                               round(float(pixel_center[1]), 3))

                # sample_mask = torch.all(pixel_ind == this_pixel_ind, dim = 1)
                sample_mask = self.pixel_receptive_field(pixel_coord, track)
                pixel_sample_current_series = self.point_sample_pixel_current(pixel_coord,
                                                                              track,
                                                                              sample_mask)
                pixel_current_series = self.compose_pixel_currents(pixel_sample_current_series, this_coarse_hit)
                pixel_timeseries[(cell_tpc,
                                  pixel_coord[0],
                                  pixel_coord[1],
                                  cell_trigger_t)] = pixel_current_series
                                           
        return pixel_timeseries

class GAMPixModel (ReadoutModel):
    """
    GAMPixModel(readout_config)

    Implementation of hit-finding logic for a GAMPix-style pixel plane.  This
    model uses a periodic measurement scheme, with a per-pixel threshold for
    digitization and recording of accumulated charge.

    Attributes
    ----------
    readout_config : ReadoutConfig object
        Config object containing specifications for tile and pixel size, gaps,
        threshold, noise, etc.
    clock_start_time : float
        timestamp of the earliest arriving charge bundle.  This serves as the
        first clock tick of the event readout.
    coarse_tile_hits : array-like[CoarseTileHit]
        Array of coarse tile hits populated by tile_hit_finding method
    fine_pixel_hits : array-like[CoarseTileHit]
        Array of pixel hits populated by pixel_hit_finding method

    See Also
    --------
    ReadoutModel : Parent class for readout models, in which the general scheme
        for charge simulation is implemented.
    LArPixModel : Sub-class implementing a LArPix-like hit-finding scheme.    
    
    """
    def tile_hit_finding(self, track, tile_timeseries, nonoise = False, **kwargs):
        """
        readout.tile_hit_finding(track,
                                 tile_timeseries,
                                 nonoise = False,
                                 **kwargs)

        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length).

        Parameters
        ----------
        track : Track object
            Input event representation containing a populated drifted track.
        tile_timeseries : array-like
            Array containing charge timeseries on a given tile.
        nonoise : bool
            Simulate front-end noise during the hit finding process.  Disable
            to simulate the mean behavior for this signal (useful for
            hypothesis testing).

        Returns
        -------
        hits : list[CoarseTileSample]
            List of found hits on the coarse tiles.
        
        """
        # TODO: fix logic for integration_length > 1
        hits = []
        
        for tile_key, timeseries in tile_timeseries.items():
            tile_tpc = tile_key[0]
            tile_center = (tile_key[1], tile_key[2])

            time_ticks, interval_charge = timeseries
            
            hold_length = self.readout_config['coarse_tiles']['integration_length']            
            
            # search along the bins until no more threshold crossings
            no_more_hits = False
            while not no_more_hits:
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]
                
                threshold = self.readout_config['coarse_tiles']['noise']*self.readout_config['coarse_tiles']['threshold_sigma']

                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0

                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]
                    
                    threshold_crossing_t = time_ticks[hit_index] 

                    # is there a better way to do this?
                    threshold_crossing_z = threshold_crossing_t*self.physics_config['charge_drift']['drift_speed']
                    
                    threshold_crossing_charge = window_charge[hit_index]
                    if not nonoise:
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['coarse_tiles']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(CoarseGridSample(tile_tpc,
                                                 tile_center,
                                                 threshold_crossing_t.item(),
                                                 threshold_crossing_z.item(),
                                                 threshold_crossing_charge.item()))
                else:
                    no_more_hits = True

        track.coarse_tiles_samples = hits
        return hits 

    def pixel_hit_finding(self, track, pixel_timeseries, nonoise = False, **kwargs):
        """
        readout.pixel_hit_finding(track,
                                  tile_timeseries,
                                  nonoise = False,
                                  **kwargs)

        This method of hit finding records the charge on a pixel
        to a capacitor buffer at each time tick.
        If the total charge collected is above a threshold,
        digitize each measurement

        Parameters
        ----------
        track : Track object
            Input event representation containing a populated drifted track.
        pixel_timeseries : array-like
            Array containing charge timeseries on a given pixel.
        nonoise : bool
            Simulate front-end noise during the hit finding process.  Disable
            to simulate the mean behavior for this signal (useful for
            hypothesis testing).

        Returns
        -------
        hits : list[PixelSample]
            List of found hits on the pixel.
        
        """
        hits = []
        
        for pixel_key, timeseries in pixel_timeseries.items():
            pixel_tpc = pixel_key[0]
            pixel_center = (pixel_key[1], pixel_key[2])
            cell_trigger_t = pixel_key[3]

            time_ticks, interval_charge = timeseries
            
            discrim_charge = torch.sum(interval_charge)+torch.poisson(torch.tensor(self.readout_config['pixels']['noise']).float())
            threshold = self.readout_config['pixels']['noise']*self.readout_config['pixels']['threshold_sigma']
            if discrim_charge > threshold:
                measured_charge = interval_charge
                if not nonoise:
                    measured_charge += torch.poisson(self.readout_config['pixels']['noise']*torch.ones_like(interval_charge))
                
                for this_timestamp, this_measured_charge in zip(time_ticks, measured_charge):
                    this_z = this_timestamp*self.physics_config['charge_drift']['drift_speed'] 

                    hits.append(PixelSample(pixel_tpc,
                                            pixel_center,
                                            this_timestamp.item(),
                                            this_z.item(),
                                            this_measured_charge.item()))
        track.pixel_samples = hits
        return hits 

class LArPixModel (ReadoutModel):
    """
    LArPixModel(readout_config)

    Implementation of hit-finding logic for a LArPix-style pixel plane.  This
    model uses a threshold and buffer measurement scheme, where charge is allowed
    to accumulate for a fixed number of clock cycles after threshold crossing
    before digitization and flushing.

    Attributes
    ----------
    readout_config : ReadoutConfig object
        Config object containing specifications for tile and pixel size, gaps,
        threshold, noise, etc.
    clock_start_time : float
        timestamp of the earliest arriving charge bundle.  This serves as the
        first clock tick of the event readout.
    coarse_tile_hits : array-like[CoarseTileHit]
        Array of coarse tile hits populated by tile_hit_finding method
    fine_pixel_hits : array-like[CoarseTileHit]
        Array of pixel hits populated by pixel_hit_finding method

    See Also
    --------
    ReadoutModel : Parent class for readout models, in which the general scheme
        for charge simulation is implemented.
    GAMPixModel : Sub-class implementing a GAMPix-like hit-finding scheme.
    
    """
    def tile_hit_finding(self, track, tile_timeseries, nonoise = False, **kwargs):
        """
        readout.tile_hit_finding(track,
                                 tile_timeseries,
                                 nonoise = False,
                                 **kwargs)

        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length)
        
        Parameters
        ----------
        track : Track object
            Input event representation containing a populated drifted track.
        tile_timeseries : array-like
            Array containing charge timeseries on a given tile.
        nonoise : bool
            Simulate front-end noise during the hit finding process.  Disable
            to simulate the mean behavior for this signal (useful for
            hypothesis testing).

        Returns
        -------
        hits : list[CoarseTileSample]
            List of found hits on the coarse tiles.
        
        """
        # TODO: fix logic for integration_length > 1
        hits = []
        
        for tile_center, timeseries in tile_timeseries.items():
            time_ticks, interval_charge = timeseries
            
            hold_length = self.readout_config['coarse_tiles']['integration_length']            
            
            # search along the bins until no more threshold crossings
            no_more_hits = False
            while not no_more_hits:
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]
                
                threshold = self.readout_config['coarse_tiles']['noise']*self.readout_config['coarse_tiles']['threshold_sigma']

                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0

                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]
                    
                    threshold_crossing_t = time_ticks[hit_index]

                    # is there a better way to do this?
                    threshold_crossing_z = threshold_crossing_t*self.physics_config['charge_drift']['drift_speed'] 
                
                    threshold_crossing_charge = window_charge[hit_index]
                    if not nonoise:
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['coarse_tiles']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(CoarseGridSample(tile_center,
                                                 threshold_crossing_t.item(),
                                                 threshold_crossing_z.item(),
                                                 threshold_crossing_charge.item()))
                else:
                    no_more_hits = True

        track.coarse_tiles_samples = hits
        return hits 

    def pixel_hit_finding(self, track, pixel_timeseries, nonoise = False, **kwargs):
        """
        readout.pixel_hit_finding(track,
                                  tile_timeseries,
                                  nonoise = False,
                                  **kwargs)

        This method of hit finding simply looks for a quantity
        of charge above threshold within a given z-bin
        (corresponding to a clock_period*integration_length)
        
        Parameters
        ----------
        track : Track object
            Input event representation containing a populated drifted track.
        pixel_timeseries : array-like
            Array containing charge timeseries on a given pixel.
        nonoise : bool
            Simulate front-end noise during the hit finding process.  Disable
            to simulate the mean behavior for this signal (useful for
            hypothesis testing).

        Returns
        -------
        hits : list[PixelSample]
            List of found hits on the pixel.
        
        """
        hits = []
        
        for pixel_key, timeseries in pixel_timeseries.items():
            pixel_center = (pixel_key[0], pixel_key[1])
            cell_trigger_t = pixel_key[2]

            time_ticks, interval_charge = timeseries

            hold_length = self.readout_config['pixels']['integration_length']            

            # search along the bins until no more threshold crossings
            no_hits = False
            while not no_hits:
                window_charge = torch.conv_tbc(interval_charge[:,None,None],
                                               torch.ones(hold_length,1,1),
                                               bias = torch.zeros(1),
                                               pad = hold_length-1)[:,0,0]
                window_charge = window_charge[hold_length-1:]

                threshold = self.readout_config['pixels']['noise']*self.readout_config['pixels']['threshold_sigma']
                threshold_crossing_mask = window_charge > threshold
                threshold_crossing_mask *= interval_charge > 0
                
                if torch.any(threshold_crossing_mask):
                    hit_index = threshold_crossing_mask.nonzero()[0][0]

                    threshold_crossing_t = time_ticks[hit_index]

                    # is there a better way to do this?
                    threshold_crossing_z = threshold_crossing_t*self.physics_config['charge_drift']['drift_speed'] 

                    threshold_crossing_charge = window_charge[hit_index]

                    if not nonoise:
                        # add quiescent noise
                        threshold_crossing_charge += torch.poisson(torch.tensor(self.readout_config['pixels']['noise']).float())

                    interval_charge[:hit_index+hold_length] = 0

                    hits.append(PixelSample(pixel_center,
                                            threshold_crossing_t.item(),
                                            threshold_crossing_z.item(),
                                            threshold_crossing_charge.item()))
                else:
                    no_hits = True

        track.pixel_samples = hits
        return hits 

class DetectorModel:
    """
    DetectorModel(detector_params,
                  physics_params,
                  readout_params)

    Detector model class.  This class defines the overall flow of simulation for
    charges within the detector volume.

    Attributes
    ----------
    detector_params : DetectorConfig
    physics_params : PhysicsConfig
    readout_params : ReadoutConfig
    readout_model : ReadoutModel object
    
    """
    def __init__(self,
                 detector_params = default_detector_params,
                 physics_params = default_physics_params,
                 readout_params = default_readout_params):

        self.detector_params = detector_params
        self.physics_params = physics_params
        self.readout_params = readout_params

        self.coordinate_manager = CoordinateManager(detector_params)
        
        self.readout_model = GAMPixModel(readout_config = readout_params,
                                         physics_config = physics_params,
                                         detector_config = detector_params)
        # self.readout_model = LArPixModel(readout_params)
 
    def simulate(self, track, **kwargs):
        """
        detector.simulate(track, **kwargs)

        Perform drift and readout simulations for the input Track object.

        Parameters
        ----------
        track : Track obect
            Event data provided by an input parser or a generator.
        
        """
        self.drift(track, **kwargs)
        self.readout(track, **kwargs)
        
    def drift(self, sampled_track, **kwargs):
        """
        detector.drift(sampled_track, **kwargs)

        Drift the charge samples from their input position
        to the anode position as defined by detector_params.
        Save the drifted positions to the track object.

        Parameters
        ----------
        track : Track obect
            Event data provided by an input parser or a generator.
        
        """
        # first, translate the raw track position into tpc coordinates
        # this translates from (x, y, z) in experimental system to
        # (i_TPC, x, y, z) in the coordinate system of the appropriate TPC
        # in the case when there are overlapping drift volumes (this should
        # be avoided), those charges will have an entry for each volume
        self.coordinate_manager.generate_tpc_coords(sampled_track)
        
        # z-component is the distance to the anode
        drift_time = sampled_track.tpc_track['position'][:,2]/self.physics_params['charge_drift']['drift_speed'] # s

        # D according to https://lar.bnl.gov/properties/trans.html#diffusion-l
        # sigma = sqrt(2*D*t)

        # use the nominal drift time to calculate diffusion
        # then, add the appropriate arrival time dispersion later
        sigma_transverse = torch.sqrt(2*self.physics_params['charge_drift']['diffusion_transverse']*drift_time)
        sigma_longitudinal = torch.sqrt(2*self.physics_params['charge_drift']['diffusion_longitudinal']*drift_time)

        diffusion_sigma = torch.stack((sigma_transverse,
                                       sigma_transverse,
                                       sigma_longitudinal,
                                       )).T

        drifted_positions = torch.normal(sampled_track.tpc_track['position'],
                                         diffusion_sigma)

        # charge is diminished by attenuation
        drifted_charges = sampled_track.tpc_track['charge']*torch.exp(-drift_time/self.physics_params['charge_drift']['electron_lifetime'])

        # add dispersion to the arrival of charge due to longitudinal diffusion
        time_dispersion = (drifted_positions[:, 2] - sampled_track.tpc_track['position'][:, 2])/self.physics_params['charge_drift']['drift_speed'] 
        
        arrival_time = drift_time + sampled_track.tpc_track['time'] + time_dispersion
            
        # might also include a sub-sampling step?
        # in case initial sampling is not fine enough

        sampled_track.drifted_track = {'position': drifted_positions,
                                       'charge': drifted_charges,
                                       'time': arrival_time}
        return 

    def readout(self, drifted_track, **kwargs):
        """
        detector.drift(drifted_track, **kwargs)

        Apply the readout simulation to the drifted track.  This is defined
        by the readout_model attribute.
        
        Parameters
        ----------
        drifted_track : Track obect
            Event data provided by an input parser or a generator.  This object
            should have already had drift simulation applied so that drifted_track
            attribute is defined and non-empty.
        
        """
        self.readout_model.electronics_simulation(drifted_track, **kwargs)

        return



from gampixpy.units import *

class MobilityModel:
    """
    MobilityModel(physics_config)

    Return a model object which can be called to compute mobility parameters
    (charge mobility, diffusion constants, etc.) from a physics parameter
    specification.

    Parameters
    ----------
    physics_config : PhysicsConfig
        A dict-like config object for physics params.

    Returns
    -------
    out : dict
        A dict object containing computed mobility parameters. This can be used
        to update the physics config (as it is called by the PhysicsConfig init
        method), or on its own.

    Examples
    --------
    >>> physics_config = gampixpy.config.default_phsyics_params
    >>> mobility_model = MobilityModel(physics_config)
    >>> mobility_model.compute_parameters()
    {'electron_mobility': 320.22680490595025,
     'drift_energy': 2.069477981671871e-08,
     'diffusion_longitudinal': 6.627023218939979e-06,
     'diffusion_ratio': 0.5005312849756938,
     'diffusion_transverse': 1.323997803506287e-05,
     'drift_speed': 0.16011340245297512}

    """
    def __init__(self, physics_config):
        self.physics_config = physics_config

    def compute_parameters(self):
        # compute drift velocity, diffusion model

        # Using the mobility model described in https://arxiv.org/abs/1508.07059
        drift_params = self.physics_config['charge_drift']
        
        a0 = drift_params['mobility_parameter_a0'] # cm^2/V/s
        a1 = drift_params['mobility_parameter_a1']
        a2 = drift_params['mobility_parameter_a2']
        a3 = drift_params['mobility_parameter_a3']
        a4 = drift_params['mobility_parameter_a4']
        a5 = drift_params['mobility_parameter_a5']
        T0 = drift_params['mobility_reference_T0']

        b0 = drift_params['mobility_parameter_b0']
        b1 = drift_params['mobility_parameter_b1']
        b2 = drift_params['mobility_parameter_b2']
        b3 = drift_params['mobility_parameter_b3']
        T1 = drift_params['mobility_reference_T1']

        E = drift_params['drift_field']
        T = self.physics_config['material']['temperature']

        numerator = a0 + a1*E/(kV/cm) + a2*pow(E/(kV/cm), 3./2) + a3*pow(E/(kV/cm), 5./2)
        denominator = 1 + (a1/a0)*E/(kV/cm) + a4*pow(E/(kV/cm), 2) + a5*pow(E/(kV/cm), 3)
        mu = (numerator*pow(T/T0, -3./2)/denominator)*cm*cm/V/s
        
        dMudE = ((a1 + 3./2*a2*pow(E/(kV/cm), 1./2) + 5./2*a3*pow(E/(kV/cm), 3./2))*denominator - numerator*(a1/a0 + 2*a4*E/(kV/cm) + 3*a5*pow(E/(kV/cm), 2)))*pow(T/T0, -3./2)/pow(denominator, 2)*cm*cm*cm/V/kV/s

        drift_energy = (b0 + b1*E/(kV/cm) + b2*pow(E/(kV/cm), 2))*(T/T1)/(1 + (b1/b0)*E/(kV/cm) + b3*pow(E/(kV/cm), 2))*eV

        R = 1 + E/mu*dMudE

        computed_params = {}
        computed_params['electron_mobility'] = mu
        computed_params['drift_energy'] = drift_energy
        computed_params['diffusion_longitudinal'] = mu*drift_energy/e
        computed_params['diffusion_ratio'] = R
        computed_params['diffusion_transverse'] = mu*drift_energy/R/e
        computed_params['drift_speed'] = mu*E

        return computed_params



from gampixpy.tracks import Track
from gampixpy.config import default_physics_params, default_detector_params
from gampixpy.units import *
from gampixpy.recombination import BoxRecombinationModel, BirksRecombinationModel

import numpy as np
import torch
import particle

meta_dtype =  np.dtype([("event id", "u4"),
                        ("primary energy", "f4"),
                        ("deposited charge", "f4"),
                        ("vertex x", "f4"),
                        ("vertex y", "f4"),
                        ("vertex z", "f4"),
                        ("theta", "f4"),
                        ("phi", "f4"),
                        ("primary length", "f4"),
                        ],
                       align = True)

class InputParser:
    """
    InputParser(input_filename,
                sequential_sampling = True,
                physics_config = default_physics_params)

    Parent class for more specialized input parsers.

    Attributes
    ----------
    physics_config : PhysicsConfig object
        Physics configuration.  Some early physics processes are handled
        by the input parser (for now!), such as charge/light yield
        calculation.
    input_filename : str or os.path-like
        Path to the input data on disk.
    sampling_order : array-like
        Array describing the order for iterating through the input indices.
    """
    def __init__(self, input_filename,
                 position_offset = [0, 0, 0],
                 sequential_sampling = True,
                 physics_config = default_physics_params,
                 detector_config = default_detector_params):

        self.physics_config = physics_config
        self.detector_config = detector_config

        self.input_filename = input_filename

        self.sampling_order = []
        self.global_position_offset = torch.tensor(position_offset)

        self._open_file_handle()
        self._generate_sample_order(sequential_sampling)

    def __len__(self):
        if 'n_images' in dir(self):
            return self.n_images
        else:
            return None
        
    def __iter__(self):
        for sample_index in self.sampling_order:
            yield sample_index.item(), self.get_sample(sample_index.item()), self.get_meta(sample_index.item())

class SegmentParser (InputParser):
    """
    SegmentParser(input_filename,
                sequential_sampling = True,
                physics_config = default_physics_params)

    Parent class for segment-based input parsers.  This class defines
    methods for point sampline along segment-like inputs and recombination.

    Attributes
    ----------
    physics_config : PhysicsConfig object
        Physics configuration.  Some early physics processes are handled
        by the input parser (for now!), such as charge/light yield
        calculation.
    input_filename : str or os.path-like
        Path to the input data on disk.
    sampling_order : array-like
        Array describing the order for iterating through the input indices.
    
    """
    def do_recombination(self, dE, dx, dEdx, mode = 'birks', **kwargs):
        """
        parser.do_recombination(dE, dx, dEdx, mode = 'birks', **kwargs)

        Calculate the recombination factor from input segment attributes.

        Parameters
        ----------
        dE : array-like[float]
            Energy deposition (integrated) over the input segment.
        dx : array-like[float]
            Length of each input segment.
        dEdx : array-like[float]
            Energy-deposition density (average) over the input segment.
            
        mode : string
            Recombination model to choose from.  Currently implemented are
            'box' and 'birks'.

        Returns
        -------
        n_electrons : array-like[float]
            Charge yield in number of electrons for each segment described
            in the input arrays.

        See Also
        --------
        recombination.BoxRecombinationModel : Implementation of the Box
            recombination model.
        recombination.BirksRecombinationModel : Implementation of the Birks
            recombination model.
        
        """
        if mode.lower() == 'box':
            recombination_model = BoxRecombinationModel(self.physics_config)
        elif mode.lower() == 'birks':
            recombination_model = BirksRecombinationModel(self.physics_config)
        else:
            raise ValueError("Invalid recombination mode: must be 'physics.BOX' or 'physics.BRIKS'")

        recomb = recombination_model(dE, dx, dEdx)

        if torch.any(torch.isnan(recomb)):
            raise RuntimeError("Invalid recombination value")

        w_ion = self.physics_config['material']['w']
        charge_yield_per_energy = recomb/w_ion

        n_electrons = dE*charge_yield_per_energy
        return n_electrons
    
    def do_point_sampling(self, start_4vec, end_4vec,
                          dx, charge_per_segment,
                          sample_density = 1.e-1,
                          sample_normalization = 'charge',
                          **kwargs):
        """
        parser.do_point_sampling(start_4vec, end_4vec,
                                 dx, charge_per_segment,
                                 sample_density = 1.e-1,
                                 sample_normalization = 'charge',
                                 **kwargs)

        Sample points along the segments described by the input arrays.

        Parameters
        ----------
        start_4vec : array-like
            (position, time) 4-vector array for the initial points of
            the image segments.
        end_4vec : array-like
            (position, time) 4-vector array for the final points of
            the image segments.
        dx : array-like
            Length array for the image segments.
        charge_per_segment : array-like
            Charge per segment array for the image segments.
        sample_density : float
            Density of samples.  Exact interpretation depends upon the
            normalization method specified.
        sample_normalization : str
            Normalization method for choosing sample density:

                'charge' : Using this method, sample_density has units
                    of [samples/e].  The samples will therefore also
                    have charge e/sample_density.
                'length' : Using this method, sample_density has units
                    of [samples/cm].  This method has the potential to
                    ignore extremely small segments with high ionization.
        
        Returns
        -------
        sample_position : array-like[float, float, float]
            Interpolated position along the input segments.
        sample_time : array-like[float]
            Interpolated ionization time along the input segments.
        charges : array-like[float]
            Charge values for each interpolated point.
        
        """

        segment_interval = end_4vec - start_4vec

        if sample_normalization == 'charge':
            # here, sample_density is [samples/unit charge]
            samples_per_segment = (charge_per_segment*sample_density).int()
        elif sample_normalization == 'length':
            # here, sample_density is [samples/unit length]
            samples_per_segment = (dx*sample_density).int()

        sample_start = torch.repeat_interleave(start_4vec, samples_per_segment, dim = 0)
        sample_interval = torch.repeat_interleave(segment_interval, samples_per_segment, dim = 0)
        if torch.any(samples_per_segment):
            sample_parametric_distance = torch.cat(tuple(torch.linspace(0, 1, samples_per_segment[i])
                                                         for i in range(samples_per_segment.shape[0])))
            sample_4vec = sample_start + sample_interval*sample_parametric_distance[:,None]
        else:
            sample_4vec = torch.empty((0, 4))
            
        sample_position = sample_4vec[:,:3]
        sample_time = sample_4vec[:,3]
        sample_charge = torch.repeat_interleave(charge_per_segment/samples_per_segment, samples_per_segment)

        return sample_position, sample_time, sample_charge

class RooTrackerParser (SegmentParser):
    """
    RooTrackerParser(input_filename,
                     sequential_sampling = True,
                     physics_config = default_physics_params)

    Class for parsing inputs from ROOT (edep-sim output).

    Attributes
    ----------
    physics_config : PhysicsConfig object
        Physics configuration.  Some early physics processes are handled
        by the input parser (for now!), such as charge/light yield
        calculation.
    input_filename : str or os.path-like
        Path to the input data on disk.
    sampling_order : array-like
        Array describing the order for iterating through the input indices.
    
    """
    def _open_file_handle(self, **kwargs):
        from ROOT import TFile, TG4Event

        self.file_handle = TFile(self.input_filename)
        self.inputTree = self.file_handle.Get("EDepSimEvents")

        self.event = TG4Event()
        self.inputTree.SetBranchAddress("Event", self.event)

    def _generate_sample_order(self, sequential_sampling, **kwargs):
        self.n_images = self.inputTree.GetEntriesFast()
        if sequential_sampling:
            self.sampling_order = torch.arange(self.n_images)
        else:
            self.sampling_order = torch.randperm(self.n_images)

    def _get_G4_segments(self, sample_index, position_offset = None, **kwargs):
        self.inputTree.GetEntry(sample_index, **kwargs)

        traj_id = torch.empty((0))
        traj_pdg = torch.empty((0))
        for trajectory in self.event.Trajectories:
            this_traj_id = torch.tensor([trajectory.GetTrackId()])
            this_traj_pdg = torch.tensor([trajectory.GetPDGCode()])

            traj_id = torch.cat((this_traj_id,
                                 traj_id))
            traj_pdg = torch.cat((this_traj_pdg,
                                 traj_pdg))
        
        start_pos = torch.empty((0,4))
        end_pos = torch.empty((0,4))
        start_time = torch.empty((0))
        end_time = torch.empty((0))
        dE = torch.empty((0))
        pdgid = torch.empty((0))

        for container_name, hit_segments in self.event.SegmentDetectors:
            for segment in hit_segments:

                this_start_pos = torch.tensor([segment.GetStart().X()*mm,
                                                segment.GetStart().Y()*mm,
                                                segment.GetStart().Z()*mm])
                this_end_pos = torch.tensor([segment.GetStop().X()*mm,
                                              segment.GetStop().Y()*mm,
                                              segment.GetStop().Z()*mm])

                this_start_time = torch.tensor([segment.GetStart().T()*ns])
                this_end_time = torch.tensor([segment.GetStop().T()*ns])

                this_dE = torch.tensor([segment.GetEnergyDeposit()*MeV])

                parent_id = list(segment.Contrib)[0]
                this_pdgid = traj_pdg[traj_id == parent_id]
                
                start_pos = torch.cat((this_start_pos[None,:],
                                       start_pos))
                end_pos = torch.cat((this_end_pos[None,:],
                                      end_pos))
                start_time = torch.cat((this_start_time,
                                        start_time))
                end_time = torch.cat((this_end_time,
                                      end_time))
                dE = torch.cat((this_dE,
                                dE))
                pdgid = torch.cat((this_pdgid,
                                   pdgid))

        return start_pos, end_pos, start_time, end_time, dE, pdgid
            
    def _get_G4_sample(self, sample_index, position_offset = None, **kwargs):
        self.inputTree.GetEntry(sample_index, **kwargs)

        start_4vec = torch.empty((0,4))
        end_4vec = torch.empty((0,4))
        dE = torch.empty((0))

        for container_name, hit_segments in self.event.SegmentDetectors:
            for segment in hit_segments:

                this_start_4vec = torch.tensor([segment.GetStart().X()*mm,
                                                segment.GetStart().Y()*mm,
                                                segment.GetStart().Z()*mm,
                                                segment.GetStart().T()*ns])
                this_end_4vec = torch.tensor([segment.GetStop().X()*mm,
                                              segment.GetStop().Y()*mm,
                                              segment.GetStop().Z()*mm,
                                              segment.GetStop().T()*ns])
                this_dE = torch.tensor([segment.GetEnergyDeposit()*MeV])
                
                start_4vec = torch.cat((this_start_4vec[None,:],
                                        start_4vec))
                end_4vec = torch.cat((this_end_4vec[None,:],
                                      end_4vec))
                dE = torch.cat((this_dE,
                                dE))

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sqrt(torch.sum(displacement**2, dim = 1))
        dEdx = torch.where(dx > 0, dE/dx, 0.)

        dQ = self.do_recombination(dE, dx, dEdx, **kwargs)
        charge_position, charge_time, charge_values = self.do_point_sampling(start_4vec,
                                                                             end_4vec,
                                                                             dx, dQ,
                                                                             **kwargs
                                                                             )

        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
            
        return Track(charge_position + offset,
                     charge_time,
                     charge_values)

    def _get_G4_meta(self, sample_index,
                     position_offset = None,
                     **kwargs):
        primary_vertex = self.event.Primaries[0] # assume only one primary for now

        offset = np.array(self.global_position_offset.cpu())
        if position_offset:
            offset += np.array(position_offset)

        vertex_x = primary_vertex.GetPosition().X()*mm + offset[0]
        vertex_y = primary_vertex.GetPosition().Y()*mm + offset[1]
        vertex_z = primary_vertex.GetPosition().Z()*mm + offset[2]
        
        primary_trajectory = self.event.Trajectories[0]
        assert primary_trajectory.GetParentId() == -1

        pdg_code = primary_trajectory.GetPDGCode()
        mass = particle.Particle.from_pdgid(pdg_code).mass*MeV
        energy = primary_trajectory.GetInitialMomentum()[3]*MeV
        momentum = [primary_trajectory.GetInitialMomentum()[i]*MeV for i in range(3)]
        kinetic_energy = energy - mass

        theta = np.arctan2(momentum[1], momentum[0])
        phi = np.arctan2(np.sqrt(momentum[0]**2 + momentum[1]**2), momentum[2])
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                -1, # charge undefined
                                vertex_x, vertex_y, vertex_z,
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_segments(self, index, **kwargs):
        """
        parser.get_segments(index, **kwargs)

        Get the N Geant4 segments associated with this event index from the
        input file.  These are normally passed directly to the point sampling
        method when `get_sample` is called.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        start_pos : array-like
            (N, 3) array containing the start points for each segment in the event.
        end_pos : array-like
            (N, 3) array containing the final points for each segment in the event.
        start_time : array-like
            (N,) array containing the time of the initial point in each segment in
            the event.
        end_time : array-like
            (N,) array containing the time of the final point in each segment in the
            event.
        dE : array-like
            (N,) array containing the energy deposited via ionizations for each
            segment in the event.
        pdgid : array-like
            (N,) array containing the PDG code for each segment in the event.
                
        """
        return self._get_G4_segments(index, **kwargs)

    def get_sample(self, index, **kwargs):
        """
        parser.get_sample(index, **kwargs)

        Get the sample image from the loaded file.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        sample : Track object
            Return the loaded image as a point-sampled track object.
        
        """
        return self._get_G4_sample(index, **kwargs)

    def get_meta(self, index, **kwargs):
        """
        parser.get_meta(index, **kwargs)

        Get the metadata from the loaded file

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        meta_array : array-like[float]
            Return an array containing the metadata for this sample.  The dtype
            of this array is defined in meta_dtype, above.
        
        """
        return self._get_G4_meta(index, **kwargs)

class EdepSimParser (SegmentParser):
    # Unit conventions for edepsim inputs:
    # distance: cm
    # energy: MeV
    def _open_file_handle(self, **kwargs):
        import h5py
        self.file_handle = h5py.File(self.input_filename, **kwargs)

    def _generate_sample_order(self, sequential_sampling, **kwargs):
        unique_event_ids = np.unique(self.file_handle['trajectories']['event_id']).astype(np.int32)
        # unique_event_ids = torch.tensor(unique_event_ids, dtype = torch.int32)
        self.n_images = len(unique_event_ids)
        if sequential_sampling:
            self.sampling_order = torch.tensor(unique_event_ids)
        else:
            self.sampling_order = torch.tensor(unique_event_ids[torch.randperm(self.n_images)])

    def _get_edepsim_segments(self, sample_index,
                              pdg_selection=None,
                              position_offset = None,
                              **kwargs):
        segment_mask = self.file_handle['segments']['event_id'] == sample_index
        if type(pdg_selection) in [np.ndarray, list]:
            pdg_selection = np.array(pdg_selection)
            segment_mask *= np.sum(self.file_handle['segments']['pdg_id'] == pdg_selection[:,None],
                                   axis = 0, dtype = bool)
        elif pdg_selection:
            segment_mask *= self.file_handle['segments']['pdg_id'] == pdg_selection
        event_segments = self.file_handle['segments'][segment_mask]
      
        trajectory_mask = self.file_handle['trajectories']['event_id'] == sample_index
        event_trajectories = self.file_handle['trajectories'][trajectory_mask]

        start_pos = np.array([event_segments['x_start']*cm,
                              event_segments['y_start']*cm,
                              event_segments['z_start']*cm]).T
        start_pos = torch.tensor(start_pos)
        end_pos = np.array([event_segments['x_end']*cm,
                            event_segments['y_end']*cm,
                            event_segments['z_end']*cm]).T
        end_pos = torch.tensor(end_pos)
        
        start_time = torch.tensor(event_segments['t_start']*ns)
        end_time = torch.tensor(event_segments['t_end']*ns)
        
        dE = torch.tensor(event_segments['dE']*MeV)
        pdgid = torch.tensor(event_segments['pdg_id'])

        return start_pos, end_pos, start_time, end_time, dE, pdgid
            
    def _get_edepsim_event(self, sample_index,
                           pdg_selection=None,
                           position_offset = None,
                           **kwargs):
        segment_mask = self.file_handle['segments']['event_id'] == sample_index
        if type(pdg_selection) in [np.ndarray, list]:
            pdg_selection = np.array(pdg_selection)
            segment_mask *= np.sum(self.file_handle['segments']['pdg_id'] == pdg_selection[:,None],
                                   axis = 0, dtype = bool)
        elif pdg_selection:
            segment_mask *= self.file_handle['segments']['pdg_id'] == pdg_selection
        event_segments = self.file_handle['segments'][segment_mask]
      
        trajectory_mask = self.file_handle['trajectories']['event_id'] == sample_index
        event_trajectories = self.file_handle['trajectories'][trajectory_mask]

        start_4vec = np.array([event_segments['x_start']*cm,
                               event_segments['y_start']*cm,
                               event_segments['z_start']*cm,
                               event_segments['t_start']*ns,
                               ]).T
        start_4vec = torch.tensor(start_4vec)
        end_4vec = np.array([event_segments['x_end']*cm,
                             event_segments['y_end']*cm,
                             event_segments['z_end']*cm,
                             event_segments['t_end']*ns,
                             ]).T
        end_4vec = torch.tensor(end_4vec)
        dE = torch.tensor(event_segments['dE']*MeV)

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sqrt(torch.sum(displacement**2, dim = 1))
        dEdx = torch.where(dx > 0, dE/dx, 0.)

        dQ = self.do_recombination(dE, dx, dEdx, **kwargs)
        charge_position, charge_time, charge_values = self.do_point_sampling(start_4vec,
                                                                             end_4vec,
                                                                             dx, dQ,
                                                                             **kwargs
                                                                             )
        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
            
        return Track(charge_position + offset,
                     charge_time,
                     charge_values)
    
    def _get_edepsim_meta(self, sample_index,
                          position_offset = None,
                          **kwargs):
        trajectory_mask = self.file_handle['trajectories']['event_id'] == sample_index
        event_trajectories = self.file_handle['trajectories'][trajectory_mask]
        primary_trajectory_mask = event_trajectories['parent_id'] == -1
        primary_trajectory_mask *= event_trajectories['traj_id'] == 0
        primary_trajectory = event_trajectories[primary_trajectory_mask]
        
        pdg_code = primary_trajectory['pdg_id']
        mass = particle.Particle.from_pdgid(pdg_code).mass*MeV
        momentum = primary_trajectory['pxyz_start']*MeV
        kinetic_energy = np.sqrt(np.power(mass, 2) + np.sum(np.power(momentum, 2))) - mass

        offset = np.array(self.global_position_offset.cpu())
        if position_offset:
            offset += np.array(position_offset)

        vertex = primary_trajectory['xyz_start'][0] + offset
        init_momentum = primary_trajectory['pxyz_start'][0]

        theta = np.arctan2(init_momentum[1], init_momentum[0])
        phi = np.arctan2(np.sqrt(init_momentum[0]**2 + init_momentum[1]**2), init_momentum[2])
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                -1, # charge undefined
                                vertex[0], vertex[1], vertex[2],
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_segments(self, index, **kwargs):
        """
        parser.get_segments(index, **kwargs)

        Get the N Geant4 segments associated with this event index from the
        input file.  These are normally passed directly to the point sampling
        method when `get_sample` is called.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        start_pos : array-like
            (N, 3) array containing the start points for each segment in the event.
        end_pos : array-like
            (N, 3) array containing the final points for each segment in the event.
        start_time : array-like
            (N,) array containing the time of the initial point in each segment in
            the event.
        end_time : array-like
            (N,) array containing the time of the final point in each segment in the
            event.
        dE : array-like
            (N,) array containing the energy deposited via ionizations for each
            segment in the event.
        pdgid : array-like
            (N,) array containing the PDG code for each segment in the event.
                
        """
        return self._get_edepsim_segments(index, **kwargs)
        
    def get_sample(self, index, **kwargs):
        """
        parser.get_sample(index, **kwargs)

        Get the sample image from the loaded file.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        sample : Track object
            Return the loaded image as a point-sampled track object.
        
        """
        return self._get_edepsim_event(index, **kwargs)

    def get_meta(self, index, **kwargs):
        """
        parser.get_meta(index, **kwargs)

        Get the metadata from the loaded file

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        meta_array : array-like[float]
            Return an array containing the metadata for this sample.  The dtype
            of this array is defined in meta_dtype, above.
        
        """
        return self._get_edepsim_meta(index, **kwargs)

class MarleyParser (SegmentParser):
    # BROKEN
    # revisit this later
    def _open_file_handle(self):
        from ROOT import TFile

        self.file_handle = TFile(self.input_filename)
        self.inputTree = self.file_handle.Get("edep")

    def _generate_sample_order(self, sequential_sampling):
        event_ids = []
        for entry in self.inputTree:
            event_ids.append(entry.event)
        unique_event_ids = np.unique(event_ids)
        self.n_images = self.inputTree.GetEntriesFast()
        if sequential_sampling:
            self.sampling_order = unique_event_ids
        else:
            self.sampling_order = np.random.choice(unique_event_ids,
                                                   len(unique_event_ids),
                                                   replace = False)
    
    def _get_G4_sample(self, sample_index,
                       position_offset = None,
                       **kwargs,
                       ):
        segment_dtype = np.dtype([("x_start", "f4"),
                                  ("y_start", "f4"),
                                  ("z_start", "f4"),
                                  ("x_end", "f4"),
                                  ("y_end", "f4"),
                                  ("z_end", "f4"),
                                  ("dE", "f4"),
                                  ("dx", "f4"),
                                  ("dEdx", "f4")],
                                align = True)
        segment_array = np.array([], dtype = segment_dtype)

        for entry in self.inputTree:
            if entry.event == sample_index:
                x_start = entry.startX
                y_start = entry.startY
                z_start = entry.startZ
                x_end = entry.endX
                y_end = entry.endY
                z_end = entry.endZ

                x_d = x_end - x_start
                y_d = y_end - y_start
                z_d = z_end - z_start
                dx = np.sqrt(x_d**2 + y_d**2 + z_d**2)
                
                dE = entry.dE
                dEdx = dE/dx if dx > 0 else 0

                this_segment_array = np.array([(x_start,
                                                y_start,
                                                z_start,
                                                x_end,
                                                y_end,
                                                z_end,
                                                dE,
                                                dx,
                                                dEdx)],
                                              dtype = segment_dtype)
                segment_array = np.concatenate((this_segment_array,
                                                segment_array))

        charge_per_segment = self.do_recombination(segment_array)
        charge_position, charge_time, charge_values = self.do_point_sampling(segment_array, charge_per_segment)

        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
            
        return Track(charge_position + offset,
                     charge_time,
                     charge_values)

    def _get_G4_meta(self, sample_index,
                     **kwargs):
        meta_array = np.array([(sample_index,
                                -1, # KE undefined
                                -1, # charge undefined
                                -1, -1, -1, # vertex point undefined
                                -1, -1, # primary attitude undefined
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index, **kwargs):
        """
        parser.get_sample(index, **kwargs)

        Get the sample image from the loaded file.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        sample : Track object
            Return the loaded image as a point-sampled track object.
        
        """
        return self.get_G4_sample(index, **kwargs)

    def get_meta(self, index, **kwargs):
        """
        parser.get_meta(index, **kwargs)

        Get the metadata from the loaded file

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        meta_array : array-like[float]
            Return an array containing the metadata for this sample.  The dtype
            of this array is defined in meta_dtype, above.
        
        """
        return self.get_G4_meta(index, **kwargs)

class MarleyCSVParser (SegmentParser):
    def _open_file_handle(self):
        self.col_names = ('run', 'subrun', 'event',
                          'isSignal', 'pdgCode', 'trackID',
                          'motherID', 'startE', 'dE',
                          'startX', 'startY',
                          'startZ', 'startT',
                          'endX', 'endY',
                          'endZ', 'endT',
                          )
        self.col_types = ('i4', 'i4', 'i4',
                          'b1', 'i4', 'i4',
                          'i4', 'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          'f4', 'f4',
                          )

        self.data_table = np.loadtxt(self.input_filename,
                                     skiprows = 1, # first row is a header column
                                     delimiter = ',', 
                                     dtype = {'names': self.col_names,
                                              'formats': self.col_types},
                                     )

    def _generate_sample_order(self, sequential_sampling):

        unique_event_ids = np.unique(self.data_table['event']).astype(np.int32)
        if sequential_sampling:
            self.sampling_order = torch.tensor(unique_event_ids)
        else:
            self.sampling_order = torch.tensor(unique_event_ids[torch.randperm(len(unique_event_ids))])

    def _get_CSV_sample(self, sample_index,
                        position_offset = None,
                        **kwargs,
                        ):
        event_mask = self.data_table['event'] == sample_index
        event_rows = self.data_table[event_mask]
        
        segment_dtype = np.dtype([("x_start", "f4"),
                                  ("y_start", "f4"),
                                  ("z_start", "f4"),
                                  ("t_start", "f4"),
                                  ("x_end", "f4"),
                                  ("y_end", "f4"),
                                  ("z_end", "f4"),
                                  ("t_end", "f4"),
                                  ("dE", "f4"),
                                  ("dx", "f4"),
                                  ("dEdx", "f4")],
                                 align = True)

        start_4vec = torch.tensor((event_rows['startY']*mm,
                                   event_rows['startZ']*mm,
                                   event_rows['startX']*mm,
                                   event_rows['startT']*ns)).T
        end_4vec = torch.tensor((event_rows['endY']*mm,
                                 event_rows['endZ']*mm,
                                 event_rows['endX']*mm,
                                 event_rows['endT']*ns)).T
        
        dE = torch.tensor(event_rows['dE']*MeV)

        displacement = start_4vec[:,:3] - end_4vec[:,:3]
        dx = torch.sqrt(torch.sum(displacement**2, dim = 1))
        dEdx = torch.where(dx > 0, dE/dx, 0.)

        dQ = self.do_recombination(dE, dx, dEdx)
        charge_position, charge_time, charge_values = self.do_point_sampling(start_4vec,
                                                                             end_4vec,
                                                                             dx, dQ,
                                                                             )
        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
            
        return Track(charge_position + offset,
                     charge_time,
                     charge_values)

    def _get_CSV_meta(self, sample_index,
                      position_offset = None,
                      **kwargs):
        event_mask = self.data_table['event'] == sample_index
        event_rows = self.data_table[event_mask]

        kinetic_energy = -1
        charge = -1
        vertex = [-1, -1, -1]
        theta = -1
        phi = -1

        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
        
        meta_array = np.array([(sample_index,
                                kinetic_energy,
                                charge, # charge undefined
                                vertex[0] + offset[0],
                                vertex[1] + offset[1],
                                vertex[2] + offset[2],
                                theta, phi,
                                -1, # primary length undefined
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index, **kwargs):
        """
        parser.get_sample(index, **kwargs)

        Get the sample image from the loaded file.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        sample : Track object
            Return the loaded image as a point-sampled track object.
        
        """
        return self._get_CSV_sample(index, **kwargs)

    def get_meta(self, index, **kwargs):
        """
        parser.get_meta(index, **kwargs)

        Get the metadata from the loaded file

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        meta_array : array-like[float]
            Return an array containing the metadata for this sample.  The dtype
            of this array is defined in meta_dtype, above.
        
        """
        return self._get_CSV_meta(index, **kwargs)

class PenelopeParser (InputParser):
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    
    def _open_file_handle(self, **kwargs):
        self.file_handle = np.load(self.input_filename, **kwargs)

    def _generate_sample_order(self, sequential_sampling):
        # These inputs have only one event per file
        # Sampling is trivial
        self.sampling_order = [0]
        
    def _get_penelope_sample(self, position_offset = None, **kwargs):
        charge_position = torch.tensor(self.file_handle['r']).T
        charge_values = torch.tensor(self.file_handle['num_e'])
        charge_time = torch.zeros_like(charge_values)
        
        offset = self.global_position_offset
        if position_offset:
            offset += torch.tensor(position_offset)
            
        return Track(charge_position + offset,
                     charge_time,
                     charge_values)

    def _get_penelope_meta(self,
                           **kwargs):
        # is the only available metadata for these energy in the filename?
        primary_energy = self.input_filename.split('/')[-1].split('_')[0][6:]
        try:
            primary_energy = float(primary_energy)*MeV
        except ValueError:
            primary_energy = -1
        
        meta_array = np.array([(0, # sample index (undefined)
                                primary_energy, # primary kinetic energy
                                -1, # charge (undefined)
                                0, 0, 0, # vertex position (undefined)
                                -1, -1, # primary attitude (undefined)
                                -1, # primary length (undefined)
                                )],
                              dtype = meta_dtype)
        return meta_array

    def get_sample(self, index, **kwargs):
        """
        parser.get_sample(index, **kwargs)

        Get the sample image from the loaded file.

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        sample : Track object
            Return the loaded image as a point-sampled track object.
        
        """
        return self._get_penelope_sample(**kwargs)
    
    def get_meta(self, index, **kwargs):
        """
        parser.get_meta(index, **kwargs)

        Get the metadata from the loaded file

        Parameters
        ----------
        index : int
            Index (in the file's internal scheme) of the sample to retrieve.

        Returns
        -------
        meta_array : array-like[float]
            Return an array containing the metadata for this sample.  The dtype
            of this array is defined in meta_dtype, above.
        
        """
        return self._get_penelope_meta()

parser_dict = {'root': RooTrackerParser,
               'edepsim': EdepSimParser,
               'marley': MarleyParser,
               'penelope': PenelopeParser,
               }



import numpy as np
import torch

from gampixpy import tracks
from gampixpy.input_parsing import meta_dtype

class Generator:
    """
    Generator(*args, **kwargs)

    Parent class for generator objects.  These are designed to work
    similarly to InputParser objects, i.e. present an iterator for
    event data and metadata arrays.

    Parameters
    ----------
    None

    Notes
    -----
    All args and kwargs to this class are saved as attributes, to
    generalize inputs for subclasses.

    See Also
    --------
    PointSource : Sub-class for generating tracks with a point profile.
    LineSource : Sub-class for generating line-like tracks with a
                 uniform charge density.
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        while True:
            yield self.get_sample(), self.get_meta()
    
class PointSource (Generator):
    """
    PointSource(xrange=None,
                yrange=None,
                zrange=None,
                trange=None,
                qrange=None)

    A generator for point sources uniformly distributed within a
    rectangular volume.  The keyword arguments specify the boundaries
    of the region from which a point is randomly sampled, and a track
    object with a preset number of points (1e5) is generated.

    Attributes
    ----------
    xrange : tuple(float, float)
        Range of x values from which to sample
    yrange : tuple(float, float)
        Range of y values from which to sample
    zrange : tuple(float, float)
        Range of z values from which to sample
    trange : tuple(float, float)
        Range of t values from which to sample
    qrange : tuple(float, float)
        Range of q values from which to sample

    Examples
    --------
    >>> ps_gen = PointSource(x_range=(0, 1),
                             y_range=(0, 1),
                             z_range=(0, 1),
                             t_range=(0, 1),
                             q_range=(0, 1))
    >>> point_cloud = ps_gen.get_sample()
    >>> meta = ps_gen.get_meta()
    >>> detector_model.simulate(point_cloud)
    
    See Also
    --------
    Generator : Parent class for generators.
    LineSource : Sub-class for generating line-like tracks with a
                 uniform charge density.
    
    """
    def generate_sample_params(self):
        """
        gen.generate_sample_params()

        Draw a new sample from the distribution parameters.

        Parameters
        ----------
        None
        
        """
        self.n_samples_per_point = 100000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.t_init = self.kwargs['t_range'][0] + (self.kwargs['t_range'][1] - self.kwargs['t_range'][0])*np.random.random()
        self.q_init = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

    def get_sample(self):
        """
        gen.get_sample()

        Generate a new sample from the parametrized distribution and return
        a Track object.
        
        Parameters
        ----------
        None

        Returns
        -------
        track : Track object
            An instance of the Track class where the sample points lie on the
            same point described in the sample parameters with uniform charge
            density.
        
        """
        
        self.generate_sample_params()

        charge_4vec = torch.tensor(self.n_samples_per_point*[[self.x_init,
                                                              self.y_init,
                                                              self.z_init,
                                                              self.t_init,
                                                              ]])
        charge_values = torch.tensor(self.n_samples_per_point*[self.q_init/self.n_samples_per_point,
                                                               ])                                 
        
        return tracks.Track(charge_4vec, charge_values)

    def get_meta(self):
        """
        gen.get_meta()

        Return the metadata associated with the current set of
        sample parameters.

        Parameters
        ----------
        None

        Returns
        -------
        meta_array : numpy.array
            An array containing metadata for this sample.  The
            dtype for this array is:

                0    "event id"            "u4"    (undefined)
                1    "primary energy"      "f4"    (undefined)
                2    "deposited charge"    "f4"
                3    "vertex x"            "f4"
                4    "vertex y"            "f4"
                5    "vertex z"            "f4"
                6    "theta"               "f4"    (undefined)
                7    "phi"                 "f4"    (undefined)
                8    "primary length"      "f4"    (undefined)

            as described in gampixpy.input_parsing.meta_dtype

        """
        
        meta_array = np.array([(0, 0,
                                self.q_init,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                0,
                                0,
                                0)],
                              dtype = meta_dtype)
                
        return meta_array 

class LineSource (Generator):
    """
    LineSource(xrange=None,
               yrange=None,
               zrange=None,
               trange=None,
               qrange=None,
               lenght_range=None)

    A generator for line sources uniformly distributed within a
    rectangular volume.  The keyword arguments specify the boundaries
    of the region from which a point is randomly sampled, and a track
    object with a preset number of points (1e5) is generated.

    Attributes
    ----------
    xrange : tuple(float, float)
        Range of x values from which to sample
    yrange : tuple(float, float)
        Range of y values from which to sample
    zrange : tuple(float, float)
        Range of z values from which to sample
    trange : tuple(float, float)
        Range of t values from which to sample
    qrange : tuple(float, float)
        Range of q values from which to sample
    length_range : tuple(float, float)
        Range of tracklet length values from which to sample

    Examples
    --------
    >>> ls_gen = LineSource(x_range=(0, 1),
                            y_range=(0, 1),
                            z_range=(0, 1),
                            t_range=(0, 1),
                            q_range=(0, 1),
                            length_range=(1,2))
    >>> line_cloud = ls_gen.get_sample()
    >>> meta = ls_gen.get_meta()
    >>> detector_model.simulate(line_cloud)
    
    See Also
    --------
    Generator : Parent class for generators.
    PointSource : Sub-class for generating tracks with a point profile.
    
    """

    def generate_sample_params(self):
        """
        gen.generate_sample_params()

        Sample from the distribution parameters.

        Parameters
        ----------
        None
        
        """
        self.n_samples = 10000

        self.x_init = self.kwargs['x_range'][0] + (self.kwargs['x_range'][1] - self.kwargs['x_range'][0])*np.random.random()
        self.y_init = self.kwargs['y_range'][0] + (self.kwargs['y_range'][1] - self.kwargs['y_range'][0])*np.random.random()
        self.z_init = self.kwargs['z_range'][0] + (self.kwargs['z_range'][1] - self.kwargs['z_range'][0])*np.random.random()
        self.t_init = self.kwargs['t_range'][0] + (self.kwargs['t_range'][1] - self.kwargs['t_range'][0])*np.random.random()
        self.q = self.kwargs['q_range'][0] + (self.kwargs['q_range'][1] - self.kwargs['q_range'][0])*np.random.random()

        # maybe direction parameters can be passed as kwargs?
        # for now, let's just always use a uniform distribution
        # over the unit sphere
        self.theta = 2*torch.pi*torch.rand(1)
        self.phi = torch.arccos(1 - 2*torch.rand(1))

        self.length = self.kwargs['length_range'][0] + (self.kwargs['length_range'][1] - self.kwargs['length_range'][0])*torch.rand(1)

    def do_point_sampling(self, start_4vec, end_4vec,
                          dx, charge_per_sample):
        """
        gen.do_point_sampling(start_4vec, end_4vec,
                              dx, charge_per_sample)

        Sample points with uniform density along a track described
        by two endpoints.

        Parameters
        ----------
        start_4vec : array-like[float, float, float, float]
            (position, time) vector for the initial point of the line
            segment.
        end_4vec : array-like[float, float, float, float]
            (position, time) vector for the final point of the line
            segment.
        dx : float
            Length of the line segment.
        charge_per_sample : float
            Charge on each point sample within the line distribution.
            Should be equivalent to total charge/n_samples

        Returns
        -------
        sample_position : array-like[float, float, float]
            Position vector for each point within the line distribution.
        sample_time : array-like[float, float, float, float]
            Time for each point within the line distribution.
        sample_charges : array-like[float]
            charge per point sample
        
        """
        # point sampling with a fixed number of samples per length
        # it may be faster to do sampling another way (test in future!)
        #  - sample with fixed amount of charge
        #  - sample with fixed number of samples per segment

        segment_interval = end_4vec - start_4vec

        sample_parametric_distance = torch.linspace(0, 1, self.n_samples)
        sample_4vec = start_4vec + segment_interval*sample_parametric_distance[:,None]

        sample_position = sample_4vec[:,:3]
        sample_time = sample_4vec[:,3]
        sample_charges = charge_per_sample*torch.ones(self.n_samples)

        return sample_position, sample_time, sample_charges
        
    def get_sample(self):
        """
        gen.get_sample()

        Generate a new sample from the parametrized distribution and return
        a Track object.
        
        Parameters
        ----------
        None

        Returns
        -------
        track : Track object
            An instance of the Track class where the sample points lie along
            the line described in the sample parameters with uniform charge
            density.
        
        """
        self.generate_sample_params()

        start_4vec = torch.tensor((self.x_init,
                                   self.y_init,
                                   self.z_init,
                                   self.t_init,
                                   ))
        dir_4vec = torch.tensor([torch.cos(torch.tensor(self.theta))*torch.sin(torch.tensor(self.phi)),
                                 torch.sin(torch.tensor(self.theta))*torch.sin(torch.tensor(self.phi)),
                                 torch.cos(torch.tensor(self.phi)),
                                 0,
                                 ])
        end_4vec = start_4vec + dir_4vec*self.length

        displacement = start_4vec[:3] - end_4vec[:3]
        dx = torch.sum(displacement**2)
        dQ = self.q/self.n_samples
        charge_4vec, charge_values = self.do_point_sampling(start_4vec,
                                                            end_4vec,
                                                            dx, dQ,
                                                            )

        return tracks.Track(charge_4vec, charge_values)

    def get_meta(self):
        """
        gen.get_meta()

        Return the metadata associated with the current set of
        sample parameters.

        Parameters
        ----------
        None

        Returns
        -------
        meta_array : numpy.array
            An array containing metadata for this sample.  The
            dtype for this array is:

                0    "event id"            "u4"    (undefined)
                1    "primary energy"      "f4"    (undefined)
                2    "deposited charge"    "f4"
                3    "vertex x"            "f4"
                4    "vertex y"            "f4"
                5    "vertex z"            "f4"
                6    "theta"               "f4"
                7    "phi"                 "f4"
                8    "primary length"      "f4"

            as described in gampixpy.input_parsing.meta_dtype

        """

        meta_array = np.array([(0, 0,
                                self.q,
                                self.x_init,
                                self.y_init,
                                self.z_init,
                                self.theta,
                                self.phi,
                                self.length)],
                              dtype = meta_dtype)
                
        return meta_array 



import numpy as np

coarse_tile_dtype = np.dtype([("event id", "u4"),
                              ("tile tpc", "u4"),
                              ("tile x", "f4"),
                              ("tile y", "f4"),
                              ("hit z", "f4"),
                              ("hit t", "f4"),
                              ("hit charge", "f4"),
                              ],
                             align = True)
pixel_dtype = np.dtype([("event id", "u4"),
                        ("pixel tpc", "u4"),
                        ("pixel x", "f4"),
                        ("pixel y", "f4"),
                        ("hit z", "f4"),
                        ("hit t", "f4"),
                        ("hit charge", "f4"),
                        ],
                       align = True)

class PixelSample:
    """
    PixelSample(pixel_tpc,
                pixel_pos,
                hit_timestamp,
                hit_depth,
                hit_measurement)

    Data container class for pixel samples.

    Attributes
    ----------
    pixel_tpc : int
        TPC index of pixel.
    pixel_pos : tuple(float, float)
        Position in anode coordinates (x, y) of pixel center.
    hit_timestamp : float
        Timestamp associated with hit.  Depending on the hit finding
        method used, this may be the time of theshold crossing or the
        time of digitization.
    hit_depth : float
        Estimated depth assiciated with this hit.  This is usually just
        arrival_time*v_drift, and so ignores some details of hit finding.
    hit_measurement : float
        Measured charge (or correlate) for this hit.
    """
    def __init__(self,
                 pixel_tpc,
                 pixel_pos,
                 hit_timestamp,
                 hit_depth,
                 hit_measurement):
        self.pixel_tpc = pixel_tpc
        self.pixel_pos = pixel_pos
        self.hit_timestamp = hit_timestamp
        self.hit_depth = hit_depth
        self.hit_measurement = hit_measurement

class CoarseGridSample:
    """
    CoarseGridSample(coarse_cell_tpc,
                     coarse_cell_pos,
                     hit_timestamp,
                     hit_depth,
                     hit_measurement)

    Data container class for coarse tile samples.

    Attributes
    ----------
    coarse_cell_tpc : int
        TPC index of coarse cell.
    coarse_cell_pos : tuple(float, float)
        Position in anode coordinates (x, y) of the tile center.
    coarse_measurement_time : float
        Timestamp associated with hit.  Depending on the hit finding
        method used, this may be the time of theshold crossing or the
        time of digitization.
    measurement_depth : float
        Estimated depth assiciated with this hit.  This is usually just
        arrival_time*v_drift, and so ignores some details of hit finding.
    coarse_cell_measurement : float
        Measured charge (or correlate) for this hit.
    """
    def __init__(self,
                 coarse_cell_tpc,
                 coarse_cell_pos,
                 measurement_time,
                 measurement_depth,
                 coarse_cell_measurement):
        self.coarse_cell_tpc = coarse_cell_tpc
        self.coarse_cell_pos = coarse_cell_pos
        self.coarse_measurement_time = measurement_time
        self.coarse_measurement_depth = measurement_depth
        self.coarse_cell_measurement = coarse_cell_measurement



from gampixpy.coordinates import CoordinateManager

import math
import numpy as np
import torch
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

import SLACplots

class EventDisplay:
    """
    EventDisplay(track)

    An event display object is a general-purpose manager for building
    3D plots of a Track object in various states of simulation.

    Parameters
    ----------
    track : Track object
        The general data container object used by gampixpy.  This may
        be at any stage of simulation.

    Attributes
    ----------
    track_object : Track
        Internal pointer to the input track object.
    fig : matplotlib.figure.Figure object
        Figure object.
    ax : matplotlib.axes._axes.Axes object
        Current (3D) axes of the figure.
    coarse_tile_hit_kwargs : dict
        Keyword args for plotting coarse hit boxes.
    pixel_hit_kwargs : dict
        Keyword args for plotting pixel hit boxes.

    Examples
    --------
    >>> evd = EventDisplay(event_data)
    >>> evd.plot_drifted_track_timeline(alpha = 0) # can also pass kwargs to plt.scatter
    >>> evd.plot_coarse_tile_measurement_timeline(readout_config) # plot tile hits
    >>> evd.plot_pixel_measurement_timeline(readout_config) # plot pixel hits
    >>> evd.show()

    """
    MAX_POINTS_PLOTTED = 1e3
    def __init__(self, track):
        self.track_object = track
        self._init_fig()

        # self.coarse_tile_hit_kwargs = dict(facecolors = 'cyan',
        self.coarse_tile_hit_kwargs = dict(facecolors=SLACplots.stanford.palo_verde,
                                           linewidths=1,
                                           edgecolors=SLACplots.SLACblue,
                                           alpha = 0.25)
        self.pixel_hit_kwargs = dict(facecolors=SLACplots.stanford.illuminating,
                                     linewidths=1,
                                     edgecolors=SLACplots.stanford.illuminating,
                                     alpha = 0.5)
        self.TPC_boundary_kwargs = dict(facecolors=None,
                                        linewidths=1,
                                        edgecolors=SLACplots.stanford.full_palette['Black']['50%'],
                                        linestyle = '--',
                                        alpha = 0)
        
    def _init_fig(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection = '3d')

    def remove_guidelines(self):
        """
        evd.remove_guidelines()

        Removes the axis grid lines, ticks, and labels.

        Parameters
        ----------
        None

        """
        self.ax.axis('off')
        
    def equal_aspect(self):
        """
        evd.equal_aspect()

        Set the aspect ratio of the plot so that x and y (anode dimensions)
        are visually equal in the axis.

        Parameters
        ----------
        None

        """
        extents = np.array([getattr(self.ax, 'get_{}lim'.format(dim))() for dim in 'xy'])
        sz = extents[:,1] - extents[:,0]
        centers = np.mean(extents, axis=1)
        maxsize = max(abs(sz))
        r = maxsize/2
        for ctr, dim in zip(centers, 'xy'):
            getattr(self.ax, 'set_{}lim'.format(dim))(ctr - r, ctr + r)
            
    def show(self):
        """
        evd.show()

        Display the plot, if there is a graphical backend and an attached display.

        Parameters
        ----------
        None

        """
        self.equal_aspect()
        plt.show()
        
    def save(self, outfile, **kwargs):
        """
        evd.save(outfile, **kwargs)

        Save the plot to disk.

        Parameters
        ----------
        outfile : str or os.path-like
            Location on disk to write the output file.

        Notes
        -----
        Additional kwargs are passed through to self.fig.savefig

        """
        self.equal_aspect()
        self.fig.savefig(outfile, **kwargs)

    def draw_box(self, cell_center_xy, cell_center_z, cell_pitch, cell_height, **kwargs):
        """
        evd.draw_box(cell_center_xy,
                     cell_center_z,
                     cell_pitch,
                     cell_height,
                     **kwargs)

        Draw a 3D box on the axes.

        Parameters
        ----------
        cell_center_xy : tuple(float, float)
            Position of box center in x and y.
        cell_center_z : float
            Position of box center in z.
        cell_pitch : float
            Extent of box in x and y directions.
        cell_height : float
            Extent of box in z direction.

        Notes
        -----
        Additional kwargs are passed through to Poly3DCollection

        """
        x_bounds = [cell_center_xy[0] - 0.5*cell_pitch,
                    cell_center_xy[0] + 0.5*cell_pitch]
        y_bounds = [cell_center_xy[1] - 0.5*cell_pitch,
                    cell_center_xy[1] + 0.5*cell_pitch]
        z_bounds = [cell_center_z,
                    cell_center_z + cell_height]

        corners = [[x_bounds[0], y_bounds[0], z_bounds[0]],
                   [x_bounds[0], y_bounds[1], z_bounds[0]],
                   [x_bounds[1], y_bounds[0], z_bounds[0]],
                   [x_bounds[1], y_bounds[1], z_bounds[0]],
                   [x_bounds[0], y_bounds[0], z_bounds[1]],
                   [x_bounds[0], y_bounds[1], z_bounds[1]],
                   [x_bounds[1], y_bounds[0], z_bounds[1]],
                   [x_bounds[1], y_bounds[1], z_bounds[1]],
                   ]

        self.draw_box_from_corners(corners, **kwargs)

    def draw_box_from_corners(self, corners, **kwargs):
        """
        evd.draw_box_from_bounds(corners,
                                 **kwargs)

        Draw a 3D box on the axes.

        Parameters
        ----------
        corners : list(array)
            A list of the eight corners of the box to be drawn.

        Notes
        -----
        Additional kwargs are passed through to Poly3DCollection

        """
        bottom_face = np.array([corners[0],
                                corners[1],
                                corners[3],
                                corners[2],
                                corners[0],
                                ])
        top_face = np.array([corners[4],
                             corners[5],
                             corners[7],
                             corners[6],
                             corners[4],
                             ])
        left_face = np.array([corners[0],
                              corners[1],
                              corners[5],
                              corners[4],
                              corners[0],
                              ])
        right_face = np.array([corners[2],
                               corners[3],
                               corners[7],
                               corners[6],
                               corners[2],
                               ])
        front_face = np.array([corners[0],
                               corners[2],
                               corners[6],
                               corners[4],
                               corners[0],
                               ])
        back_face = np.array([corners[1],
                              corners[3],
                              corners[7],
                              corners[5],
                              corners[1],
                              ])
        faces = [bottom_face,
                 top_face,
                 left_face,
                 right_face,
                 back_face,
                 front_face,
                 ]
        
        self.ax.add_collection3d(Poly3DCollection(faces, **kwargs))            

    def plot_raw_track(self, **kwargs):
        """
        evd.plot_raw_track(**kwargs)

        Plot the point sample representation of the undrifted track.

        Parameters
        ----------
        None

        Notes
        -----
        Additional kwargs are passed through to self.ax.scatter

        """

        n_points = self.track_object.raw_track['position'].shape[0]
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            xs = self.track_object.raw_track['position'][::reduction_factor,0]
            ys = self.track_object.raw_track['position'][::reduction_factor,1]
            zs = self.track_object.raw_track['position'][::reduction_factor,2]
            colors = np.log(self.track_object.raw_track['charge'][::reduction_factor])
        else:
            xs = self.track_object.raw_track['position'][:,0],
            ys = self.track_object.raw_track['position'][:,1],
            zs = self.track_object.raw_track['position'][:,2],
            colors = np.log(self.track_object.raw_track['charge'][:])
            
        self.ax.scatter(xs, ys, zs,
                        c = colors,
                        **kwargs,
                        )

        self.ax.set_xlabel(r'x [cm]')
        self.ax.set_ylabel(r'y [cm]')
        self.ax.set_zlabel(r'z [cm]')
            
    def plot_drifted_track(self, detector_config, **kwargs):
        """
        evd.plot_drifted_track(**kwargs)

        Plot the point sample representation of the drifted track.

        Parameters
        ----------
        detector_config : DetectorConfig object
            The detector config specifying the location and size of drift
            volumes in the geometry.  This should match the readout config
            used in the simulation step for this track.

        Notes
        -----
        Additional kwargs are passed through to self.ax.scatter

        """

        n_points = self.track_object.drifted_track['position'].shape[0]
        coordinate_manager = CoordinateManager(detector_config)
        coords = coordinate_manager.to_experiment_coords(self.track_object.drifted_track['position'],
                                                         self.track_object.tpc_track['TPC_index'])
        if n_points > self.MAX_POINTS_PLOTTED:
            reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
            xs = coords[::reduction_factor,0].cpu()
            ys = coords[::reduction_factor,1].cpu()
            zs = coords[::reduction_factor,2].cpu()
            colors = np.log(self.track_object.drifted_track['charge'][::reduction_factor].cpu())
        else:
            xs = coords[:,0].cpu()
            ys = coords[:,1].cpu()
            zs = coords[:,2].cpu()
            colors = np.log(self.track_object.drifted_track['charge'][:].cpu())
            
        self.ax.scatter(xs, ys, zs,
                        c = colors,
                        **kwargs,
                        )

        self.ax.set_xlabel(r'x [cm]')
        self.ax.set_ylabel(r'y [cm]')
        self.ax.set_zlabel(r'z [cm]')

    # def plot_drifted_track_timeline(self, **kwargs):
    #     """
    #     evd.plot_drifted_track_timeline(**kwargs)

    #     Plot the point sample representation of the drifted track. This method
    #     plots the arrival time on the z-axis instead of the true depth.

    #     Parameters
    #     ----------
    #     None

    #     Notes
    #     -----
    #     Additional kwargs are passed through to self.ax.scatter

    #     """

    #     n_points = self.track_object.drifted_track['position'].shape[0]
    #     if n_points > self.MAX_POINTS_PLOTTED:
    #         reduction_factor = math.ceil(n_points/self.MAX_POINTS_PLOTTED)
    #         xs = self.track_object.drifted_track['position'][::reduction_factor,0].cpu()
    #         ys = self.track_object.drifted_track['position'][::reduction_factor,1].cpu()
    #         zs = self.track_object.drifted_track['time'][::reduction_factor].cpu()
    #         colors = np.log(self.track_object.drifted_track['charge'][::reduction_factor].cpu())
    #     else:
    #         xs = self.track_object.drifted_track['position'][:,0].cpu()
    #         ys = self.track_object.drifted_track['position'][:,1].cpu()
    #         zs = self.track_object.drifted_track['time'][:].cpu()
    #         colors = np.log(self.track_object.drifted_track['charge'][:].cpu())
                        

    #     self.ax.scatter(xs, ys, zs,
    #                     c = colors,
    #                     **kwargs,
    #                     )

    #     self.ax.set_xlabel(r'x (transverse) [cm]')
    #     self.ax.set_ylabel(r'y (transverse) [cm]')
    #     self.ax.set_zlabel(r'Arrival Time [us]')
        
    def plot_coarse_tile_measurement(self, readout_config, physics_config, detector_config):
        """
        evd.plot_coarse_tile_measurement(readout_config)

        Plot the simulated coarse hits for an input track.

        Parameters
        ----------
        readout_config : ReadoutConfig object
            The readout config specifying the coarse tile pitch, clock
            interval, etc.  This should match the readout config used in
            the simulation step for this track.
        physics_config : PhysicsConfig object
            The physics config specifying physics parameters. This should
            match the physics config used in the simulation step for this
            track.
        detector_config : DetectorConfig object
            The detector config specifying the location and size of drift
            volumes in the geometry.  This should match the readout config
            used in the simulation step for this track.

        """

        coordinate_manager = CoordinateManager(detector_config)
        
        for this_hit in self.track_object.coarse_tiles_samples:
            cell_tpc = this_hit.coarse_cell_tpc
            
            cell_center_xy = this_hit.coarse_cell_pos
            cell_center_z = this_hit.coarse_measurement_depth

            tpc_coords = torch.tensor([cell_center_xy[0],
                                       cell_center_xy[1],
                                       cell_center_z])
            exp_coords = coordinate_manager.to_experiment_coords(tpc_coords,
                                                                 cell_tpc)
            exp_coords = exp_coords.cpu().numpy()[0]
            
            cell_measurement = this_hit.coarse_cell_measurement

            pitch = readout_config['coarse_tiles']['pitch'],

            this_volume_dict = detector_config['drift_volumes'][coordinate_manager.index_to_volume[cell_tpc]]
            horizontal_axis = this_volume_dict['anode_horizontal'].cpu().numpy()
            half_span_horizontal = horizontal_axis*pitch/2

            vertical_axis = this_volume_dict['anode_vertical'].cpu().numpy()
            half_span_vertical = vertical_axis*pitch/2

            drift_axis = this_volume_dict['drift_axis'].cpu().numpy()
            v = physics_config['charge_drift']['drift_speed']
            cell_hit_length = v*readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']
            depth_span = drift_axis*cell_hit_length

            corners = [exp_coords - half_span_horizontal - half_span_vertical,
                       exp_coords - half_span_horizontal + half_span_vertical,
                       exp_coords + half_span_horizontal - half_span_vertical,
                       exp_coords + half_span_horizontal + half_span_vertical,
                       exp_coords - half_span_horizontal - half_span_vertical + depth_span,
                       exp_coords - half_span_horizontal + half_span_vertical + depth_span,
                       exp_coords + half_span_horizontal - half_span_vertical + depth_span,
                       exp_coords + half_span_horizontal + half_span_vertical + depth_span,
                       ]

            self.draw_box_from_corners(corners,
                                       **self.coarse_tile_hit_kwargs)
            
        self.ax.set_xlabel(r'x [cm]')
        self.ax.set_ylabel(r'y [cm]')
        self.ax.set_zlabel(r'z [cm]')

    # def plot_coarse_tile_measurement_timeline(self, readout_config):
    #     """
    #     evd.plot_coarse_tile_measurement_timeline(readout_config)

    #     Plot the simulated coarse hits for an input track. This method
    #     plots the arrival time on the z-axis instead of the true depth.

    #     Parameters
    #     ----------
    #     readout_config : ReadoutConfig object
    #         The readout config specifying the coarse tile pitch, clock
    #         interval, etc.  This should match the readout config used in
    #         the simulation step for this track.

    #     """

    #     xlim = []
    #     ylim = []
    #     zlim = []
        
    #     for this_hit in self.track_object.coarse_tiles_samples:
    #         cell_center_xy = this_hit.coarse_cell_pos
    #         # cell_center_z = this_hit.coarse_measurement_time
    #         cell_trigger_t = this_hit.coarse_measurement_time
    #         cell_measurement = this_hit.coarse_cell_measurement
    #         cell_hit_length = readout_config['coarse_tiles']['clock_interval']*readout_config['coarse_tiles']['integration_length']

    #         self.draw_box(cell_center_xy, cell_trigger_t,
    #                       readout_config['coarse_tiles']['pitch'],
    #                       cell_hit_length,
    #                       **self.coarse_tile_hit_kwargs)
            
    #     self.ax.set_xlabel(r'x [cm]')
    #     self.ax.set_ylabel(r'y [cm]')
    #     self.ax.set_zlabel(r'Arrival Time [us]')
        
    def plot_pixel_measurement(self, readout_config, physics_config, detector_config):
        """
        evd.plot_pixel_measurement(readout_config)

        Plot the simulated pixel hits for an input track.

        Parameters
        ----------
        readout_config : ReadoutConfig object
            The readout config specifying the coarse tile pitch, clock
            interval, etc.  This should match the readout config used in
            the simulation step for this track.
        physics_config : PhysicsConfig object
            The physics config specifying physics parameters. This should
            match the physics config used in the simulation step for this
            track.
        detector_config : DetectorConfig object
            The detector config specifying the location and size of drift
            volumes in the geometry.  This should match the readout config
            used in the simulation step for this track.

        """

        coordinate_manager = CoordinateManager(detector_config)

        for this_hit in self.track_object.pixel_samples:
            cell_tpc = this_hit.pixel_tpc
            
            cell_center_xy = this_hit.pixel_pos
            cell_center_z = this_hit.hit_depth

            tpc_coords = torch.tensor([cell_center_xy[0],
                                       cell_center_xy[1],
                                       cell_center_z,
                                       ])

            exp_coords = coordinate_manager.to_experiment_coords(tpc_coords,
                                                                 cell_tpc)
            exp_coords = exp_coords.cpu().numpy()[0]

            cell_measurement = this_hit.hit_measurement

            pitch = readout_config['pixels']['pitch'],

            this_volume_dict = detector_config['drift_volumes'][coordinate_manager.index_to_volume[cell_tpc]]
            horizontal_axis = this_volume_dict['anode_horizontal'].cpu().numpy()
            half_span_horizontal = horizontal_axis*pitch/2

            vertical_axis = this_volume_dict['anode_vertical'].cpu().numpy()
            half_span_vertical = vertical_axis*pitch/2

            drift_axis = this_volume_dict['drift_axis'].cpu().numpy()
            v = physics_config['charge_drift']['drift_speed']
            cell_hit_length = v*readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']
            depth_span = drift_axis*cell_hit_length

            corners = [exp_coords - half_span_horizontal - half_span_vertical,
                       exp_coords - half_span_horizontal + half_span_vertical,
                       exp_coords + half_span_horizontal - half_span_vertical,
                       exp_coords + half_span_horizontal + half_span_vertical,
                       exp_coords - half_span_horizontal - half_span_vertical + depth_span,
                       exp_coords - half_span_horizontal + half_span_vertical + depth_span,
                       exp_coords + half_span_horizontal - half_span_vertical + depth_span,
                       exp_coords + half_span_horizontal + half_span_vertical + depth_span,
                       ]
            
            self.draw_box_from_corners(corners,
                                       **self.pixel_hit_kwargs)
            
        self.ax.set_xlabel(r'x [cm]')
        self.ax.set_ylabel(r'y [cm]')
        self.ax.set_zlabel(r'z [cm]')

    # def plot_pixel_measurement_timeline(self, readout_config):
    #     """
    #     evd.plot_pixel_measurement_timeline(readout_config)

    #     Plot the simulated pixel hits for an input track. This method
    #     plots the arrival time on the z-axis instead of the true depth.

    #     Parameters
    #     ----------
    #     readout_config : ReadoutConfig object
    #         The readout config specifying the coarse tile pitch, clock
    #         interval, etc.  This should match the readout config used in
    #         the simulation step for this track.

    #     """
        
    #     for this_hit in self.track_object.pixel_samples:
    #         cell_center_xy = this_hit.pixel_pos
    #         # cell_center_z = this_hit.hit_timestamp
    #         cell_trigger_t = this_hit.hit_timestamp
    #         cell_measurement = this_hit.hit_measurement
    #         cell_hit_length = readout_config['pixels']['clock_interval']*readout_config['pixels']['integration_length']
            
    #         self.draw_box(cell_center_xy, cell_trigger_t,
    #                       readout_config['pixels']['pitch'],
    #                       cell_hit_length,
    #                       **self.pixel_hit_kwargs)
            
    #     self.ax.set_xlabel(r'x (transverse) [cm]')
    #     self.ax.set_ylabel(r'y (transverse) [cm]')
    #     self.ax.set_zlabel(r'Arrival Time [us]')

    def plot_drift_volumes(self, detector_config):
        """
        evd.plot_drift_volumes(detector_config)

        Plot the boundaries of the drift volumes as specified in a
        detector configuration file.

        Parameters
        ----------
        detector_config : DetectorConfig object
            The detector config specifying the location and size of drift
            volumes in the geometry.  This should match the readout config
            used in the simulation step for this track.

        """

        for volume_name, volume_dict in detector_config['drift_volumes'].items():
            corners = volume_dict['anode_corners'] + volume_dict['cathode_corners']
            self.draw_box_from_corners(corners, **self.TPC_boundary_kwargs)

        self.ax.set_xlabel(r'x (transverse) [cm]')
        self.ax.set_ylabel(r'y (transverse) [cm]')
        self.ax.set_zlabel(r'z (drift) [cm]')



import torch

class RecombinationModel:
    """
    RecombinationModel(physics_config)

    Parent class for specific recombination models.

    See Also
    --------
    BoxRecombinationModel : Box recombination model
        (R. Acciarri et al JINST 8 (2013) P08005).
    BirksRecombinationModel : Birks recombination model
        (Amoruso, et al NIM A 523 (2004) 275).
    """
    def __init__(self, physics_config):
        self.physics_config = physics_config
        self.init_params()

    def __call__(self, *args, **kwargs):
        return self.recombination_factor(*args, **kwargs)

class BoxRecombinationModel (RecombinationModel):
    """
    BoxRecombinationModel

    Implementation of the Box recombination model
    (R. Acciarri et al JINST 8 (2013) P08005).

    See Also
    --------
    BirksRecombinationModel : Birks recombination model
        (Amoruso, et al NIM A 523 (2004) 275).

    """
    def init_params(self):
        self.E_field = self.physics_config['charge_drift']['drift_field']
        self.LAr_density = self.physics_config['material']['density']

        self.alpha = self.physics_config['box_model']['box_alpha']
        self.beta = self.physics_config['box_model']['box_beta']
        
    def recombination_factor(self, dE, dx, dEdx, **kwargs):
        """
        model.recombination_factor(dE, dx, dEdx)

        Calculate the recombination factor for an array of segments.
        
        Parameters
        ----------
        dE : array-like[float]
            Energy deposition per input segment.
        dx : array-like[float]
            Length of each input segment.
        dEdx : array-like[float]
            Energy deposition desity per input segment.

        Returns
        -------
        R : array-like[float]
            Recombination model for the given input parameters
            given the additional physics parameters supplied by
            the config.
        
        """
        xi = self.beta*dEdx/(self.E_field*self.LAr_density)
        recomb = torch.max(torch.stack([torch.zeros_like(dE),
                                        torch.log(self.alpha + xi)/xi]),
                           dim = 0)[0]

        return recomb

class BirksRecombinationModel (RecombinationModel):
    """
    BirksRecombinationModel(physics_config)

    Implementation of the Birks recombination model
    (Amoruso, et al NIM A 523 (2004) 275).

    See Also
    --------
    BoxRecombinationModel : Box recombination model
        (R. Acciarri et al JINST 8 (2013) P08005).
    
    """
    def init_params(self):
        self.E_field = self.physics_config['charge_drift']['drift_field']
        self.LAr_density = self.physics_config['material']['density']

        self.ab = self.physics_config['birks_model']['birks_ab']
        self.kb = self.physics_config['birks_model']['birks_kb']
        
    def recombination_factor(self, dE, dx, dEdx, **kwargs):
        """
        model.recombination_factor(dE, dx, dEdx)

        Calculate the recombination factor for an array of segments.
        
        Parameters
        ----------
        dE : array-like[float]
            Energy deposition per input segment.
        dx : array-like[float]
            Length of each input segment.
        dEdx : array-like[float]
            Energy deposition desity per input segment.

        Returns
        -------
        R : array-like[float]
            Recombination model for the given input parameters
            given the additional physics parameters supplied by
            the config.
        
        """
        recomb = torch.max(torch.stack([torch.zeros_like(dE),
                                        self.ab/(1 + self.kb*dEdx/(self.E_field*self.LAr_density))]),
                           dim = 0)[0]

        return recomb



import numpy as np

from gampixpy.readout_objects import coarse_tile_dtype, pixel_dtype

class Track:
    """
    Track (sample_position, sample_time, sample_charge)

    General-purpose data class for storing representations of an
    ionization event in the detector.

    Attributes
    ----------

    raw_track : dict
        Dict containing sample position vectors (key: 'position'),
        sample ionization time (key 'time'), and sample charge 
        values (key 'charge').  All are array-like.
    drifted_track : dict
        Dict containing sample position 3-vectors ('position'), sample
        arrival times ('time') and charge after attenuation ('charge').
    pixel_samples : list[CoarseGridSample]
        List of coarse tile hits found by detector simulation.
    coarse_tiles_samples : list[PixelSample]
        List of pixel hits found by detector simulation.

    See Also
    --------
    PixelSample : Data class for pixel hits.
    CoarseTileSample : Data class for tile hits.
    
    """
    def __init__(self, sample_position, sample_time, sample_charge):
        self.raw_track = {'position': sample_position,
                          'time': sample_time,
                          'charge': sample_charge}

        self.tpc_track = {}
        self.drifted_track = {}

        self.pixel_samples = []
        self.coarse_tiles_samples = []

    def to_array(self):
        """
        track.to_array()

        Generate a numpy array with the simulated hit data contained
        in this object for saving to a summary HDF5 file.

        Parameters
        ----------
        None

        Returns
        -------
        coarse_tile_sample_array : Flattened numpy.array of coarse hit
            data with dtype described in readout_objects.coarse_tile_dtype.
        pixel_sample_array : Flattened numpy.array of pixel hit data with
            dtype described in readout_objects.coarse_tile_dtype.

        """
        coarse_tile_sample_array = np.array([(0,
                                              hit.coarse_cell_tpc,
                                              hit.coarse_cell_pos[0],
                                              hit.coarse_cell_pos[1],
                                              hit.coarse_measurement_depth,
                                              hit.coarse_measurement_time,
                                              hit.coarse_cell_measurement)
                                             for hit in self.coarse_tiles_samples],
                                            dtype = coarse_tile_dtype)
        pixel_sample_array = np.array([(0,
                                        hit.pixel_tpc,
                                        hit.pixel_pos[0],
                                        hit.pixel_pos[1],
                                        hit.hit_depth,
                                        hit.hit_timestamp,
                                        hit.hit_measurement)
                                       for hit in self.pixel_samples],
                                      dtype = pixel_dtype)

        return coarse_tile_sample_array, pixel_sample_array



import torch

class CoordinateManager:
    """
    CoordinateManager()

    Class for transforming external coordinate systems into the internal
    coordinate system.

    Attributes
    ----------
    
    """
    def __init__(self, detector_config):
        self.detector_config = detector_config

        self.rotation = 0

        # dictionaries for translating from volume name stings to an internal index
        self.index_to_volume = {i: volume_name
                                for i, volume_name
                                in enumerate(self.detector_config['drift_volumes'].keys())}
        self.volume_to_index = {volume_name: i
                                for i, volume_name
                                in self.index_to_volume.items()}
        
        return
        
    def to_experiment_coords(self, coords, tpc_index):
        """
        coord_manager.to_external_coords(vector)

        Transform the input vector from the internal coordinate system
        to the external (experimenatal) coordinate system specified in
        the detector config to the internal system.

        Parameters
        ----------
        coords : array-like
            An (N,3) array of (x, y, z) coordinates in the per-TPC system
        tpc_index : array-like
            A (N,) array of TPC indices, indicating which origin is used
            for each point sample.

        Returns
        -------
        experiment_coords : array-like
            an (N,3) array of (x, y, z) coordinates in the external
            reference frame.

        See Also
        --------
        to_internal_coords : inverse function for transforming from external
            coordinates to internal coordinates.
        """

        exp_coords = torch.empty((0,3))
        for volume_name, volume_dict in self.detector_config['drift_volumes'].items():
            tpc_origin = volume_dict['anode_center']

            subset_mask = tpc_index == self.volume_to_index[volume_name]
            coord_subset = coords[subset_mask]

            exp_coord_subset = tpc_origin + \
                coord_subset[:,0,None]*volume_dict['anode_horizontal'][None,:] + \
                coord_subset[:,1,None]*volume_dict['anode_vertical'][None,:] + \
                coord_subset[:,2,None]*volume_dict['drift_axis'][None,:]

            exp_coords = torch.concatenate([exp_coords,
                                            exp_coord_subset,
                                            ])
            
        return exp_coords

    def to_tpc_coords(self, exp_coords):
        """
        coord_manager.to_internal_coords(vector)

        Transform the input vector from the external coordinate system
        specified in the detector config to the internal system.

        Parameters
        ----------
        exp_coords : array
            an (N, 3) array of (x, y, z) coordinates in the experimental
            reference frame.
        
        Returns
        -------
        tpc_coords : array
            an (N, 4) array of (i_TPC, x, y, z) coordinates in the TPC
            reference frame.

        See Also
        --------
        to_experiment_coords : inverse function for transforming from external
            coordinates to internal coordinates.
        """

        tpc_coords = torch.empty((0,4))
        for volume_name, volume_dict in self.detector_config['drift_volumes'].items():
            # choose an arbitrary corner.  Which one doesn't matter, but it should be
            # between 0 and 7 (a rectangular prism has 8 vertices)
            reference_corner_index = 0
            corners = volume_dict['corners']
            reference_corner = corners[reference_corner_index]
            connected_corners = corners[volume_dict['connectivity'][reference_corner_index]]

            leg_vec = connected_corners - reference_corner
            leg_dists = torch.linalg.norm(leg_vec, axis = 1)
            # project each point onto each leg, then normalize by the leg length
            extent = torch.inner(exp_coords - reference_corner, leg_vec)/leg_dists**2
            # the points which are within the drift volume are ones where the extent
            # along each leg is between 0 and 1
            keep_mask = torch.all(extent >= 0, axis = 1)*torch.all(extent <= 1, axis = 1)

            # project into TPC coordinates
            # the origin in each TPC is the anode_center, given in the detector config
            anode_disp = exp_coords[keep_mask] - volume_dict['anode_center']
            tpc_coords_subset = torch.stack([self.volume_to_index[volume_name]*torch.ones(anode_disp.shape[0]),
                                             torch.inner(anode_disp, volume_dict['anode_horizontal']),
                                             torch.inner(anode_disp, volume_dict['anode_vertical']),
                                             torch.inner(anode_disp, volume_dict['drift_axis'])]).T

            tpc_coords = torch.concatenate([tpc_coords,
                                            tpc_coords_subset])

        return tpc_coords

    def generate_tpc_coords(self, track):
        """
        coord_manager.generate_tpc_coords(vector)

        Transform the raw charge cloud from an input track from
        the experimental coordinate system to per-TPC coordinates.
        This function populates the 'tpc_track' attribute of the
        Track object.

        Parameters
        ----------
        track : Track object
            Event data provided by an input parser or a generator.
        
        Notes
        -----
        Populates the input track with a 'tpc_track' dict with the transformed
            coordinates and properly masked charge and timestamps.
        
        See Also
        --------
        to_experiment_coords : inverse function for transforming from external
            coordinates to internal coordinates.
        """
        
        tpc_index = torch.empty((0,), dtype = torch.uint8)
        tpc_coords = torch.empty((0,3))
        tpc_time = torch.empty((0,))
        tpc_charge = torch.empty((0,))
        for volume_name, volume_dict in self.detector_config['drift_volumes'].items():
            # choose an arbitrary corner.  Which one doesn't matter, but it should be
            # between 0 and 7 (a rectangular prism has 8 vertices)
            reference_corner_index = 0
            reference_corner = volume_dict['corners'][reference_corner_index]
            connected_corners = volume_dict['corners'][volume_dict['connectivity'][reference_corner_index]]

            leg_vec = connected_corners - reference_corner
            leg_dists = torch.linalg.norm(leg_vec, axis = 1)
            # project each point onto each leg, then normalize by the leg length
            extent = torch.inner(track.raw_track['position'] - reference_corner, leg_vec)/leg_dists**2
            # the points which are within the drift volume are ones where the extent
            # along each leg is between 0 and 1
            keep_mask = torch.all(extent >= 0, axis = 1)*torch.all(extent <= 1, axis = 1)

            # project into TPC coordinates
            # the origin in each TPC is the anode_center, given in the detector config
            anode_disp = track.raw_track['position'][keep_mask] - volume_dict['anode_center']
            tpc_coords_subset = torch.stack([torch.inner(anode_disp, volume_dict['anode_horizontal']),
                                             torch.inner(anode_disp, volume_dict['anode_vertical']),
                                             torch.inner(anode_disp, volume_dict['drift_axis'])]).T

            tpc_index_subset = self.volume_to_index[volume_name]*torch.ones(anode_disp.shape[0])
            tpc_time_subset = track.raw_track['time'][keep_mask]
            tpc_charge_subset = track.raw_track['charge'][keep_mask]
            
            tpc_index = torch.concatenate([tpc_index,
                                           tpc_index_subset])
            tpc_coords = torch.concatenate([tpc_coords,
                                            tpc_coords_subset])
            tpc_time = torch.concatenate([tpc_time,
                                          tpc_time_subset])
            tpc_charge = torch.concatenate([tpc_charge,
                                            tpc_charge_subset])

        track.tpc_track = {'TPC_index': tpc_index,
                           'position': tpc_coords,
                           'time': tpc_time,
                           'charge': tpc_charge,
                           }

    def to_tpc_indices(self, coords):
        """
        coord_manager.to_internal_coords(vector)

        Transform the input vector from the external coordinate system
        specified in the detector config to the internal system.

        Parameters
        ----------

        Returns
        -------

        See Also
        --------
        to_internal_coords : inverse function for transforming from external
            coordinates to internal coordinates.
        """
        # is this a useful function?  Implement at some point...
        return coords



#!/usr/bin/env python

import setuptools

VER = "0.1.9"

reqs = ['numpy',
        'h5py',
        'tqdm',
        # 'torch', # skip this for now.  Users must fend for themselves!
        # 'ROOT', # skip this for now.  Users must fend for themselves!
        'torchist',
        'particle',
        'pyyaml',
        ]
links = ['https://download.pytorch.org/whl/cpu']

setuptools.setup(
    name="gampixpy",
    version=VER,
    author="SLAC National Accelerator Laboratory",
    author_email="dougl215@slac.stanford.edu",
    description="Gampix Readout Simulation",
    url="https://github.com/DanielMDouglas/GAMPixPy",
    packages=setuptools.find_packages(),
    install_requires=reqs,
    package_data={"gampixpy": ["*_config/*.yaml"]},
    dependency_links=links,
    license_files=["LICENSE"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics"
    ],
    python_requires='>=3.2',
)
